import fsgs
import fsui
from fsui.extra.iconheader import IconHeader
from fsbc.application import app
from launcher.i18n import gettext
from launcher.ui.skin import LauncherTheme
from launcher.ui.widgets import CloseButton


class AboutDialog(fsui.Window):
    def __init__(self, parent):
        if fsgs.product == "OpenRetro":
            app_name = "OpenRetro Launcher"
        else:
            app_name = "FS-UAE Launcher"
        title = "{} - {}".format(gettext("About"), app_name)
        super().__init__(parent, title, minimizable=False, maximizable=False)
        self.theme = LauncherTheme.get()
        self.layout = fsui.VerticalLayout()
        self.layout.set_padding(20)

        self.icon_header = IconHeader(
            self,
            fsui.Icon("fs-uae-launcher", "pkg:launcher"),
            "{name} {version}".format(name=app_name, version=app.version),
            "Copyright © 2012-2017 Frode Solheim",
        )
        self.layout.add(self.icon_header, fill=True, margin_bottom=20)

        self.text_area = fsui.TextArea(
            self, about_message, read_only=True, font_family="monospace"
        )
        self.text_area.scroll_to_start()
        self.text_area.set_min_width(760)
        self.text_area.set_min_height(340)
        self.layout.add(self.text_area, fill=True, expand=True)

        CloseButton.add_to_layout(self, self.layout, margin_top=20)


about_message = """\n
This package is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 2 of the License, or (at your option) any later
version.

This package is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this package; if not, write to the Free Software Foundation, Inc.,
51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

The following people have translated FS-UAE Launcher into several languages: 
Cédric "Foul" Monféfoul (French), nexusle (German), Speedvicio (Italian),
grimi (Polish), Milanchez (Serbian), albconde (Spanish), Treco (Portuguese),
GoingDown (Finish), spajdr (Czech), Decypher (Turkisk).

A big thanks to everyone who have tested the software and provided valuable
feedback! Especially the encouraging members of the English Amiga Board, and
all you who have commented on the official FS-UAE web site.

FS-UAE Launcher includes icons from GNOME icon theme from the GNOME Project
(http://www.gnome.org). The GNOME icon theme is distributed under the terms
of either GNU LGPL v.3 or Creative Commons BY-SA 3.0 license.

FS-UAE Launcher includes icons from humanity-icon-theme, licensed under the
terms of the GNU General Public license.

FS-UAE includes icons from oxygen-icon-theme, licensed under the terms of the
GNU Lesser General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

The Amiga Forever icon is (probably) copyright Cloanto Italia srl.

FS-UAE Launcher includes the oyoyo library, Copyright (c) 2008 Duncan Fordyce,
The license for this library is contained in the following paragraph:
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software. THE SOFTWARE IS PROVIDED
"AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE
AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

FS-UAE Launcher may include lhafile, Copyright (c) 2010 Hidekazu Ohnishi.
All rights reserved. Redistribution and use in source and binary forms,
with or without modification, are permitted provided that the following
conditions are met:
* Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.
* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.
* Neither the name of the author nor the names of its contributors may be
  used to endorse or promote products derived from this software without
  specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

FS-UAE Launcher depends on several open source third party software packages,
including but not limited to, Python and PyQt/Pyside.
"""



import re
import json
import traceback
from wsgiref.util import setup_testing_defaults
from wsgiref.simple_server import make_server
from fsgs.FSGameSystemContext import FSGameSystemContext
from fsgs.session import Session
from fsgs.http.request import Request

"""

POST /sessions

GET /sessions
{
    "id": "00000000-0000-0000-0000-000000000000"
}

"""


class ClientError(Exception):
    def __init__(self, message, code=400):
        Exception.__init__(self, message)
        self.code = code


default_fsgs = FSGameSystemContext()


def fsgs_from_session(session_id):
    assert session_id == "default"
    return default_fsgs


def configure_session(request, session_id):
    assert session_id == "default"
    request.session_id = session_id
    request.fsgs = fsgs_from_session(session_id)


# def get_sessions_handler(request):
#     return {}


def json_list(the_list):
    return {"list_hack": the_list}


def get_platforms_handler(request):
    return json_list(
        [
            {"id": "amiga", "name": "Amiga", "type": "Home Computer"},
            {"id": "arcade", "name": "Arcade", "type": "Arcade Machines"},
            {
                "id": "cd32",
                "name": "CD32",
                # "name": "Amiga CD32",
                "type": "TV Console",
            },
            {
                "id": "cdtv",
                "name": "CDTV",
                # "name": "Commodore CDTV",
                "type": "TV Console",
            },
            {
                "id": "dos",
                "name": "DOS",
                # "name": "DOS (PC)",
                "type": "Home Computer",
            },
            {
                "id": "gba",
                "name": "Game Boy Advance",
                "type": "Handheld Console",
            },
            {"id": "nes", "name": "Nintendo (NES)", "type": "TV Console"},
            {"id": "snes", "name": "Super Nintendo", "type": "TV Console"},
        ]
    )


def get_games_handler(request):
    fsgs = FSGameSystemContext()
    # game_database = fsgs.get_game_database()
    # return {"platform:": request.params.platform}

    terms = []
    if request.params.platform:
        platform = request.params.platform
        # FIXME
        if platform == "snes":
            platform = "super-nintendo"
        elif platform == "gba":
            platform = "game-boy-advance"
        terms.append("platform:{}".format(platform))

    terms.append("t:thumb")

    database_items = fsgs.database().find_games_new(" ".join(terms))
    result = []
    for item in database_items:

        # query = "SELECT DISTINCT uuid, name, platform, year, publisher, " \
        #         "front_image, title_image, screen1_image, screen2_image, " \
        #         "screen3_image, screen4_image, screen5_image, have, path, " \
        #         "sort_key, subtitle FROM game"

        # backdrop = item[17].replace("sha1:", "")
        # if backdrop:
        #    # backdrop, hs, vs, ha, va = backdrop.split("/")

        result.append(
            {
                "uuid": item[0],
                "platform": item[2],
                "name": item[1],
                "subtitle": item[15],
                "year": item[3],
                "publisher": item[4],
                # "title_image": item[6].replace("sha1:", ""),
                # "screen1_image": item[7].replace("sha1:", ""),
                "thumb": item[16].replace("sha1:", ""),
                # "backdrop": backdrop,
                "backdrop": item[17],
            }
        )
    return json_list(result)


def get_session_handler(request, session_id):
    configure_session(request, session_id)
    return {}


def post_session_init_handler(request, session_id):
    configure_session(request, session_id)
    print(request.params.game_uuid)
    request.fsgs.load_game_by_uuid(request.params.game_uuid)

    from fsbc.application import app

    if request.params.fullscreen:
        # print("setting fullscreen to", repr(request.params.fullscreen))
        # request.fsgs.config.set("fullscreen", request.params.fullscreen)
        app.settings.set("fullscreen", request.params.fullscreen)
    if request.params.fullscreen_mode:
        app.settings.set("fullscreen_mode", request.params.fullscreen_mode)
    if request.params.window_border:
        app.settings.set("window_border", request.params.window_border)
    return {}


instances = {}


def post_session_instances_handler(request, session_id):
    configure_session(request, session_id)

    print(request.params.game_uuid)

    session = Session(request.fsgs)
    instance = session.create_instance()
    instances[instance.uuid] = instance
    instance.start()

    return {"instance_id": instance.uuid}


def get_instance_handler(request, session_id, instance_id):
    configure_session(request, session_id)
    try:
        instance = instances[instance_id]
    except KeyError:
        raise ClientError("Instance not found", code=404)
    return {"instance_id": instance.uuid, "state": instance.state}


def put_instance_handler(request, session_id, instance_id):
    configure_session(request, session_id)
    try:
        instance = instances[instance_id]
    except KeyError:
        raise ClientError("Instance not found", code=404)
    if request.params.state:
        ins = instance
        old = instance.state
        new = request.params.state
        if old == ins.STATE_INITIALIZED and new == ins.STATE_PREPARING:
            instance.state = new
        elif old == ins.STATE_READY and new == ins.STATE_RUNNING:
            instance.state = new
        elif old == ins.STATE_STOPPED and new == ins.STATE_FINALIZING:
            instance.state = new
        else:
            raise ClientError("Illegal state change")
    return {"instance_id": instance.uuid, "state": instance.state}


def delete_instance_handler(request, session_id, instance_id):
    configure_session(request, session_id)
    try:
        instance = instances[instance_id]
    except KeyError:
        raise ClientError("Instance not found", code=404)
    if instance.state != instance.STATE_FINALIZED:
        raise ClientError("Excepted FINALIZED state")
    instance.destroy()
    del instances[instance_id]
    return {}


class URLMapping:
    def __init__(self, method, path, handler):
        self.method = method
        self.path = path
        self.handler = handler

        self.param_names = []
        # quick hack
        # FIXME: must escape some literals, etc
        re_str = path

        if "{session_id}" in re_str:
            re_str = re_str.replace("{session_id}", "(?P<session_id>[^/]+)")
            self.param_names.append("session_id")
        if "{instance_id}" in re_str:
            re_str = re_str.replace("{instance_id}", "(?P<instance_id>[^/]+)")
            self.param_names.append("instance_id")

        re_str = "^" + re_str + "$"
        print(re_str)
        self.re = re.compile(re_str)

    def match(self, environ):
        path = environ["PATH_INFO"]
        # print(environ["REQUEST_METHOD"], path)
        if self.method != environ["REQUEST_METHOD"]:
            return None
        m = self.re.match(path)
        # print(m)
        if m is not None:
            return m.groupdict()
        return None

    def run_handler(self, request, kwargs):
        print("run_handler", self.path, "args =", kwargs)
        return self.handler(request, **kwargs)


url_map = [
    URLMapping("GET", "/games", get_games_handler),
    URLMapping("GET", "/platforms", get_platforms_handler),
    # URLMapping("GET", "/sessions", get_sessions_handler),
    URLMapping("GET", "/sessions/{session_id}", get_session_handler),
    URLMapping(
        "POST", "/sessions/{session_id}/init", post_session_init_handler
    ),
    URLMapping(
        "POST",
        "/sessions/{session_id}/instances",
        post_session_instances_handler,
    ),
    URLMapping(
        "GET",
        "/sessions/{session_id}/instances/{instance_id}",
        get_instance_handler,
    ),
    URLMapping(
        "PUT",
        "/sessions/{session_id}/instances/{instance_id}",
        put_instance_handler,
    ),
    URLMapping(
        "DELETE",
        "/sessions/{session_id}/instances/{instance_id}",
        delete_instance_handler,
    ),
]


def create_application():

    fsgs = FSGameSystemContext()

    # A relatively simple WSGI application. It's going to print out the
    # environment dictionary after being updated by setup_testing_defaults
    def simple_app(environ, start_response):
        # FIXME: remove use of setup_testing_defaults
        setup_testing_defaults(environ)
        environ["fsgs"] = fsgs

        for mapping in url_map:
            kwargs = mapping.match(environ)
            if kwargs is not None:
                request = Request(environ)
                try:
                    result = mapping.run_handler(request, kwargs)
                    status = "200 OK"
                    if isinstance(result, dict):
                        headers = [("Content-type", "application/json")]
                        if "list_hack" in result:
                            result = result["list_hack"]
                        data = json.dumps(result, indent=4).encode("UTF-8")
                    else:
                        raise Exception("unknown result data type")
                    # headers = [("Content-type", "application/octet-stream")]
                    # data = json.dumps(result).encode("UTF-8")
                except Exception as e:
                    traceback.print_exc()
                    status = "500 Internal Server Error"
                    headers = [("Content-type", "text/plain")]
                    data = repr(e).encode("UTF-8")
                start_response(status, headers)
                return [data]

        start_response("404 Not Found", [("Content-type", "text/plain")])
        return [b"The resource was not found"]

    return simple_app


def http_server_main():
    port = 15232
    app = create_application()
    httpd = make_server("127.0.0.1", port, app)
    print("Serving on port {port}...".format(port=port))
    httpd.serve_forever()



import threading
import time
import traceback

import fsui
from fsgs.Database import Database
from fsgs.ogd.context import SynchronizerContext
from fsgs.ogd.game_rating_synchronizer import GameRatingSynchronizer
from fsgs.ogd.lists import ListsSynchronizer
from fsgs.ogd.meta import MetaSynchronizer
from launcher.configuration_scanner import ConfigurationScanner
from launcher.filescanner import FileScanner
from launcher.gamescanner import GameScanner
from launcher.launcher_config import LauncherConfig
from launcher.launcher_settings import LauncherSettings
from launcher.launcher_signal import LauncherSignal


class Scanner:
    running = False
    stop_flag = False
    status = ("", "")
    error = ""
    paths = []
    update_game_database = False
    purge_other_dirs = False
    scan_for_files = False

    @classmethod
    def stop_check(cls):
        return cls.stop_flag

    @classmethod
    def scan_thread(cls):
        try:
            with Database.get_instance() as database:
                cls._scan_thread(database)
        except Exception as e:
            cls.error = str(e)
            traceback.print_exc()

        # else:
        #     if cls.on_status:
        #         cls.on_status(("", _("Committing data...")))
        #     database.commit()

        def on_done():
            # FIXME: these should be removed soon
            LauncherSettings.set("last_scan", str(time.time()))
            LauncherSettings.set("__config_refresh", str(time.time()))

            # this must be called from main, since callbacks are broadcast
            # when settings are changed

            LauncherSignal.broadcast("scan_done")
            LauncherConfig.update_kickstart()

        # call on_done from main thread
        fsui.call_after(on_done)
        cls.running = False

    @classmethod
    def _scan_thread(cls, database):
        if cls.scan_for_files:
            scanner = FileScanner(
                cls.paths,
                cls.purge_other_dirs,
                on_status=cls.on_status,
                stop_check=cls.stop_check,
            )
            scanner.scan()
            if cls.stop_check():
                return

        # if cls.scan_for_configs or
        # if cls.update_game_database:
        scanner = ConfigurationScanner(
            cls.paths, on_status=cls.on_status, stop_check=cls.stop_check
        )
        scanner.scan(database)

        if cls.update_game_database:

            context = SynchronizerContext()

            synchronizer = MetaSynchronizer(
                context, on_status=cls.on_status, stop_check=cls.stop_check
            )
            synchronizer.synchronize()

            synchronizer = GameRatingSynchronizer(
                context,
                database,
                on_status=cls.on_status,
                stop_check=cls.stop_check,
            )
            synchronizer.username = "auth_token"
            synchronizer.password = LauncherSettings.get("database_auth")
            synchronizer.synchronize()

            synchronizer = ListsSynchronizer(
                context, on_status=cls.on_status, stop_check=cls.stop_check
            )
            synchronizer.synchronize()

            scanner = GameScanner(
                context,
                cls.paths,
                on_status=cls.on_status,
                stop_check=cls.stop_check,
            )
            scanner.update_game_database()

        scanner = GameScanner(
            None, cls.paths, on_status=cls.on_status, stop_check=cls.stop_check
        )
        scanner.scan(database)

    @classmethod
    def start(
        cls,
        paths,
        scan_for_files=True,
        update_game_database=False,
        purge_other_dirs=False,
    ):
        print("Scanner.start")
        if cls.running:
            print("scan already in progress")
            return
        cls.paths = paths[:]
        cls.running = True
        cls.error = ""
        cls.stop_flag = False
        cls.status = ("", "")
        # cls.scan_for_roms = scan_for_roms
        cls.scan_for_files = scan_for_files
        cls.purge_other_dirs = purge_other_dirs
        # cls.scan_for_configs = scan_for_configs
        cls.update_game_database = update_game_database
        threading.Thread(target=cls.scan_thread, name="ScannerThread").start()

    @classmethod
    def on_status(cls, status):
        cls.status = status



import fsui
from fsbc import settings
from fsbc.util import unused
from fsgs import openretro
from fsgs.amiga.amiga import Amiga
from fsgs.context import fsgs
from fsgs.platform import PLATFORM_IDS
from launcher.option import Option
from fsgs.platforms import PLATFORM_ZXS
from launcher.cd_manager import CDManager
from launcher.floppy_manager import FloppyManager
from launcher.i18n import gettext
from launcher.launcher_config import LauncherConfig
from launcher.panels.additionalconfigpanel import CustomConfigButton
from launcher.ui.behaviors.amigaenablebehavior import AmigaEnableBehavior
from launcher.ui.behaviors.configbehavior import ConfigBehavior
from launcher.ui.behaviors.platformbehavior import (
    AMIGA_PLATFORMS,
    PlatformShowBehavior,
    AmigaShowBehavior,
)
from launcher.ui.config.ConfigCheckBox import ConfigCheckBox
from launcher.ui.options import ConfigWidgetFactory


class ModelGroup(fsui.Group):

    # FIXME: remove with_more_button=True
    def __init__(self, parent, with_more_button=True):
        unused(with_more_button)
        fsui.Group.__init__(self, parent)
        self.layout = fsui.VerticalLayout()

        self.model_ids = [x["id"] for x in Amiga.models if "/" not in x["id"]]
        self.model_titles = [
            x["title"] for x in Amiga.models if "/" not in x["id"]
        ]

        self.sub_model_ids = []
        self.sub_model_titles = []
        self.sub_model_updating = False

        self.model_choice = fsui.Choice(self, self.model_titles)
        # AmigaEnableBehavior(self.model_choice)
        self.sub_model_choice = fsui.Choice(self, self.sub_model_titles)
        # AmigaEnableBehavior(self.sub_model_choice)
        self.accuracy_label = fsui.Label(self, gettext("Accuracy:"))
        self.accuracy_choice = fsui.Choice(
            self, [gettext("High"), gettext("Medium"), gettext("Low")]
        )
        # AmigaEnableBehavior(self.accuracy_choice)
        self.ntsc_checkbox = ConfigCheckBox(self, "NTSC", Option.NTSC_MODE)

        AmigaShowBehavior(self.accuracy_label)
        AmigaShowBehavior(self.accuracy_choice)
        AmigaShowBehavior(self.ntsc_checkbox)

        # if fs_uae_launcher.ui.get_screen_size()[1] > 768:
        # self.layout.add(heading_label, margin=10)
        # self.layout.add_spacer(0)

        self.model_title_layout = fsui.HorizontalLayout()
        self.layout.add(self.model_title_layout, fill=True)

        if openretro or settings.get(Option.PLATFORMS_FEATURE) == "1":
            heading_label = fsui.HeadingLabel(
                self, gettext("Platform & Model")
            )
            self.model_title_layout.add(heading_label, margin=10)
            # platform_group = ConfigWidgetFactory(
            #     check=False, label=False).create(self, Option.PLATFORM)
            # self.model_title_layout.add(platform_group, margin_left=20)
            # Adding label to get the vertical spacing correct.
            # heading_label = fsui.HeadingLabel(self, "")
            # self.model_title_layout.add(heading_label, margin=10)
        else:
            heading_label = fsui.HeadingLabel(self, gettext("Amiga Model"))
            self.model_title_layout.add(heading_label, margin=10)

        self.model_title_layout.add_spacer(0, expand=True)
        self.model_title_layout.add(
            self.ntsc_checkbox, expand=False, margin_left=10, margin_right=10
        )
        self.model_title_layout.add_spacer(20)

        self.model_title_layout.add(self.accuracy_label, margin_right=10)
        self.model_title_layout.add(self.accuracy_choice, margin_right=10)
        self.model_title_layout.add(CustomConfigButton(self), margin_right=10)

        self.model_layout = fsui.HorizontalLayout()

        def dummy_min_width():
            return 0

        # Not sure why this is needed, but on startup, the min width
        # seems to be set too large due to something in the model layout.
        self.model_layout.get_min_width = dummy_min_width
        self.layout.add(self.model_layout, fill=True)

        if openretro or settings.get(Option.PLATFORMS_FEATURE) == "1":
            platform_group = ConfigWidgetFactory(
                check=False, label=False
            ).create(self, Option.PLATFORM)
            self.model_layout.add(platform_group, margin=10)
            pass

        self.other_model_choice = ModelChoice(self)
        self.model_layout.add(self.other_model_choice, expand=True, margin=10)

        self.model_layout.add(self.model_choice, expand=False, margin=10)
        AmigaShowBehavior(self.model_choice)
        self.model_layout.add(self.sub_model_choice, expand=True, margin=10)
        AmigaShowBehavior(self.sub_model_choice)

        ConfigBehavior(
            self, [Option.ACCURACY, Option.AMIGA_MODEL, Option.PLATFORM]
        )

        self.model_choice.on_changed = self.on_model_changed
        self.sub_model_choice.on_changed = self.on_sub_model_changed
        self.accuracy_choice.on_changed = self.on_accuracy_changed

    def on_platform_config(self, _):
        # Update layout after widgets have been shown/hidden.
        self.model_title_layout.update()
        self.model_layout.update()

    def on_model_changed(self):
        print("ModelGroup.on_model_change\n")
        index = self.model_choice.get_index()
        model = self.model_ids[index]
        if model == "A500":
            # The default model (A500) can be specified with the empty string
            model = ""
        LauncherConfig.set("amiga_model", model)
        # Config.update_kickstart()
        # if Amiga.is_cd_based(Config):
        #     FloppyManager.clear_all()
        # else:
        #     CDManager.clear_all()

    def on_sub_model_changed(self):
        print("ModelGroup.on_sub_model_change\n")
        if self.sub_model_updating:
            print("sub model list is currently updating")
            return
        index = self.sub_model_choice.get_index()
        # if index == 0:
        #     # The default model (A500) can be specified with the empty string
        #     model = ""
        # else:
        model = self.model_ids[self.model_choice.get_index()]
        sub_model = self.sub_model_ids[index]
        if sub_model:
            LauncherConfig.set("amiga_model", model + "/" + sub_model)
        else:
            LauncherConfig.set("amiga_model", model)

        if Amiga.is_cd_based(LauncherConfig):
            FloppyManager.clear_all()
        else:
            CDManager.clear_all()

    def on_accuracy_changed(self):
        index = self.accuracy_choice.get_index()
        if index == 0:
            LauncherConfig.set("accuracy", "")
        else:
            LauncherConfig.set("accuracy", str(1 - index))

    def update_sub_models(self, model_id, sub_model_id):
        sub_model_index = 0
        model_id_s = model_id + "/"
        self.sub_model_ids.clear()
        self.sub_model_titles.clear()

        for i, config in enumerate(Amiga.models):
            if config["id"] == model_id:
                self.sub_model_ids.append("")
                self.sub_model_titles.append(config["subtitle"])
            elif config["id"].startswith(model_id_s):
                self.sub_model_ids.append(config["id"].split("/", 1)[1])
                self.sub_model_titles.append(config["subtitle"])
            else:
                continue
            if sub_model_id == self.sub_model_ids[-1]:
                sub_model_index = len(self.sub_model_ids) - 1

        self.sub_model_choice.clear()
        for title in self.sub_model_titles:
            self.sub_model_choice.add_item(title)
        self.sub_model_choice.enable(
            LauncherConfig.get(Option.PLATFORM) in AMIGA_PLATFORMS
            and len(self.sub_model_ids) > 1
        )
        return sub_model_index

    def on_amiga_model_config(self, value):
        if value == "":
            value = "A500"

        if "/" in value:
            model_id, sub_model_id = value.split("/", 1)
        else:
            model_id = value
            sub_model_id = ""

        model_index = 0
        sub_model_index = 0
        self.sub_model_updating = True
        for i, config in enumerate(Amiga.models_config):
            if config == value:
                # self.model_choice.set_index(i)
                # find main model index
                model_index = self.model_ids.index(model_id)
                sub_model_index = self.update_sub_models(
                    model_id, sub_model_id
                )
                # model_index = i
                break
        # else:
        #    print("FIXME: could not set model")
        self.model_choice.set_index(model_index)
        self.sub_model_choice.set_index(sub_model_index)
        self.sub_model_updating = False

    def on_accuracy_config(self, value):
        if not value:
            index = 0
        else:
            index = 1 - int(value)
        self.accuracy_choice.set_index(index)


class ModelChoice(fsui.Choice):
    def __init__(self, parent):
        self._choice_values = []
        self._choice_labels = []
        super().__init__(parent, self._choice_labels)
        self._platform = ""
        self._model_key = ""
        fsgs.signal.connect("config", self.on_config)
        self.on_config(Option.PLATFORM, fsgs.config.get(Option.PLATFORM))
        self.changed.connect(self.__changed)
        self.set_min_width(100)

    def on_destroy(self):
        fsgs.signal.disconnect("config", self.on_config)

    def __changed(self):
        fsgs.config.set(self._model_key, self._choice_values[self.get_index()])

    def on_config(self, key, value):
        if key == Option.PLATFORM:
            self._platform = value
            self._model_key = value + "_model"
            self.update_options()
            self.update_index(fsgs.config.get(self._model_key))
            self.update_enabled()
        elif key == self._model_key:
            self.update_index(value)

    def update_enabled(self):
        self.set_visible(self._platform not in AMIGA_PLATFORMS)
        self.set_enabled(self._choice_labels != ["N/A"])

    def update_index(self, value):
        try:
            index = self._choice_values.index(value)
        except ValueError:
            index = 0
        with self.changed.inhibit:
            self.set_index(index)

    def update_options(self):
        try:
            option = Option.get(self._model_key)
        except KeyError:
            self._choice_values = ["0"]
            self._choice_labels = ["N/A"]
        else:
            choices = option["values"]
            self._choice_values = [x[0] for x in choices]
            self._choice_labels = [x[1] for x in choices]
        with self.changed.inhibit:
            self.clear()
            for label in self._choice_labels:
                self.add_item(label)



from fsbc.settings import Settings
from fsgs.FSGSDirectories import FSGSDirectories

"""
Debug script used to dump information about detected plugins
"""


def app_main():
    FSGSDirectories.initialize()
    Settings.instance().verbose = False
    Settings.instance().load()
    print("base_dir", FSGSDirectories.get_base_dir())
    print("cache_dir", FSGSDirectories.get_cache_dir())
    print("cdroms_dir", FSGSDirectories.get_cdroms_dir())
    print("configurations_dir", FSGSDirectories.get_configurations_dir())
    print("controllers_dir", FSGSDirectories.get_controllers_dir())
    print("covers_dir", FSGSDirectories.get_covers_dir())
    print("data_dir", FSGSDirectories.get_data_dir())
    print("downloads_dir", FSGSDirectories.downloads_dir())
    print("floppies_dir", FSGSDirectories.get_floppies_dir())
    print("hard_drives_dir", FSGSDirectories.get_hard_drives_dir())
    # print("images_dir", FSGSDirectories.get_images_dir())
    print("kickstarts_dir", FSGSDirectories.get_kickstarts_dir())
    print("launcher_dir", FSGSDirectories.get_launcher_dir())
    print("logs_dir", FSGSDirectories.get_logs_dir())
    print("plugins_dir", FSGSDirectories.get_plugins_dir())
    # print("portable_dir", FSGSDirectories.portable_dir())
    print("save_states_dir", FSGSDirectories.get_save_states_dir())
    print("screenshots_dir", FSGSDirectories.get_screenshots_dir())
    print("screenshots_output_dir =", FSGSDirectories.screenshots_output_dir())
    print("themes_dir", FSGSDirectories.get_themes_dir())
    print("titles_dir", FSGSDirectories.get_titles_dir())



"""
from dinput.h
"""

dinput_key_codes = {
    "BACKSPACE": 0x0E,
    "TAB": 0x0F,
    "CLEAR": 0,  # FIXME
    "RETURN": 0x1C,
    "PAUSE": 0,  # FIXME
    "ESCAPE": 0x01,
    "SPACE": 0x39,
    "EXCLAIM": 0,  # FIXME
    "QUOTEDBL": 0,  # FIXME
    "HASH": 0,  # FIXME
    "DOLLAR": 0,  # FIXME
    "AMPERSAND": 0,  # FIXME
    "QUOTE": 0x28,
    "LEFTPAREN": 0,  # FIXME
    "RIGHTPAREN": 0,  # FIXME
    "ASTERISK": 0,  # FIXME
    "PLUS": 0,  # FIXME
    "COMMA": 0x33,
    "MINUS": 0x0C,
    "PERIOD": 0x34,
    "SLASH": 0x35,
    "0": 0x0B,
    "1": 0x02,
    "2": 0x03,
    "3": 0x04,
    "4": 0x05,
    "5": 0x06,
    "6": 0x07,
    "7": 0x08,
    "8": 0x09,
    "9": 0x0A,
    "COLON": 0x92,
    "SEMICOLON": 0x27,
    "LESS": 0,  # FIXME
    "EQUALS": 0x0D,
    "GREATER": 0,  # FIXME
    "QUESTION": 0,  # FIXME
    "AT": 0x91,
    "LEFTBRACKET": 0x1A,
    "BACKSLASH": 0x2B,
    "RIGHTBRACKET": 0x1B,
    "CARET": 0,  # FIXME
    "UNDERSCORE": 0x93,
    "BACKQUOTE": 0,  # FIXME
    "A": 0x1E,
    "B": 0x30,
    "C": 0x2E,
    "D": 0x20,
    "E": 0x12,
    "F": 0x21,
    "G": 0x22,
    "H": 0x23,
    "I": 0x17,
    "J": 0x24,
    "K": 0x25,
    "L": 0x26,
    "M": 0x32,
    "N": 0x31,
    "O": 0x18,
    "P": 0x19,
    "Q": 0x10,
    "R": 0x13,
    "S": 0x1F,
    "T": 0x14,
    "U": 0x16,
    "V": 0x2F,
    "W": 0x11,
    "X": 0x2D,
    "Y": 0x15,
    "Z": 0x2C,
    "DELETE": 0xD3,
    "KP0": 0x52,
    "KP1": 0x4F,
    "KP2": 0x50,
    "KP3": 0x51,
    "KP4": 0x4B,
    "KP5": 0x4C,
    "KP6": 0x4D,
    "KP7": 0x47,
    "KP8": 0x48,
    "KP9": 0x49,
    "KP_PERIOD": 0x53,
    "KP_DIVIDE": 0xB5,
    "KP_MULTIPLY": 0x37,
    "KP_MINUS": 0x4A,
    "KP_PLUS": 0x4E,
    "KP_ENTER": 0x9C,
    "KP_EQUALS": 0x8D,
    "UP": 0xC8,
    "DOWN": 0xD0,
    "RIGHT": 0xCD,
    "LEFT": 0xCB,
    "INSERT": 0xD2,
    "HOME": 0xC7,
    "END": 0xCF,
    "PAGEUP": 0xC9,
    "PAGEDOWN": 0xD1,
    "F1": 0x3B,
    "F2": 0x3C,
    "F3": 0x3D,
    "F4": 0x3E,
    "F5": 0x3F,
    "F6": 0x40,
    "F7": 0x41,
    "F8": 0x42,
    "F9": 0x43,
    "F10": 0x44,
    "F11": 0x57,
    "F12": 0x58,
    "F13": 0x64,
    "F14": 0x65,
    "F15": 0x66,
    "NUMLOCK": 0x45,
    "CAPSLOCK": 0x3A,
    "SCROLLOCK": 0x46,
    "RSHIFT": 0x36,
    "LSHIFT": 0x2A,
    "RCTRL": 0x9D,
    "LCTRL": 0x1D,
    "RALT": 0xB8,
    "LALT": 0x38,
    "RMETA": 0,  # FIXME
    "LMETA": 0,  # FIXME
    "LSUPER": 0xDB,
    "RSUPER": 0xDC,
    "MODE": 0,  # FIXME
    "COMPOSE": 0,  # FIXME
    "HELP": 0,  # FIXME
    "PRINT": 0,  # FIXME
    "SYSREQ": 0xB7,
    "BREAK": 0,  # FIXME
    "MENU": 0,  # FIXME
    "POWER": 0,  # FIXME
    "EURO": 0,  # FIXME
    "UNDO": 0,  # FIXME
}


sdlk_to_dik = {
    "ESCAPE": "ESCAPE",
    "1": "1",
    "2": "2",
    "3": "3",
    "4": "4",
    "5": "5",
    "6": "6",
    "7": "7",
    "8": "8",
    "9": "9",
    "0": "0",
    "MINUS": "MINUS",
    "EQUALS": "EQUALS",
    "BACKSPACE": "BACK",
    "TAB": "TAB",
    "Q": "Q",
    "W": "W",
    "E": "E",
    "R": "R",
    "T": "T",
    "Y": "Y",
    "U": "U",
    "I": "I",
    "O": "O",
    "P": "P",
    "LEFTBRACKET": "LBRACKET",
    "RIGHTBRACKET": "RBRACKET",
    "RETURN": "RETURN",
    "LCTRL": "LCONTROL",
    "A": "A",
    "S": "S",
    "D": "D",
    "F": "F",
    "G": "G",
    "H": "H",
    "J": "J",
    "K": "K",
    "L": "L",
    "SEMICOLON": "SEMICOLON",
    "QUOTE": "APOSTROPHE",
    "BACKQUOTE": "GRAVE",
    "LSHIFT": "LSHIFT",
    "BACKSLASH": "BACKSLASH",
    "Z": "Z",
    "X": "X",
    "C": "C",
    "V": "V",
    "B": "B",
    "N": "N",
    "M": "M",
    "COMMA": "COMMA",
    "PERIOD": "PERIOD",
    "SLASH": "SLASH",
    "RSHIFT": "RSHIFT",
    "KP_MULTIPLY": "MULTIPLY",
    "LALT": "LMENU",
    "SPACE": "SPACE",
    "CAPSLOCK": "CAPITAL",
    "F1": "F1",
    "F2": "F2",
    "F3": "F3",
    "F4": "F4",
    "F5": "F5",
    "F6": "F6",
    "F7": "F7",
    "F8": "F8",
    "F9": "F9",
    "F10": "F10",
    "NUMLOCK": "NUMLOCK",
    "SCROLLOCK": "SCROLL",
    "KP7": "NUMPAD7",
    "KP8": "NUMPAD8",
    "KP9": "NUMPAD9",
    "KP_MINUS": "SUBTRACT",
    "KP4": "NUMPAD4",
    "KP5": "NUMPAD5",
    "KP6": "NUMPAD6",
    "KP_PLUS": "ADD",
    "KP1": "NUMPAD1",
    "KP2": "NUMPAD2",
    "KP3": "NUMPAD3",
    "KP0": "NUMPAD0",
    "KP_PERIOD": "DECIMAL",
    "F11": "F11",
    "F12": "F12",
    "F13": "F13",
    "F14": "F14",
    "F15": "F15",
    # "": "KANA",
    # "": "CONVERT",
    # "": "NOCONVERT",
    # "": "YEN",
    # "": "NUMPADEQUALS",
    # "": "CIRCUMFLEX",
    "AT": "AT",
    "COLON": "COLON",
    # "": "UNDERLINE",
    # "": "KANJI",
    # "": "STOP",
    # "": "AX",
    # "": "UNLABELED",
    "KP_ENTER": "NUMPADENTER",
    "RCTRL": "RCONTROL",
    # "": "NUMPADCOMMA",
    "KP_DIVIDE": "DIVIDE",
    "SYSREQ": "SYSRQ",
    "RALT": "RMENU",
    "HOME": "HOME",
    "UP": "UP",
    "PAGEUP": "PRIOR",
    "LEFT": "LEFT",
    "RIGHT": "RIGHT",
    "END": "END",
    "DOWN": "DOWN",
    "PAGEDOWN": "NEXT",
    "INSERT": "INSERT",
    "DELETE": "DELETE",
    "LSUPER": "LWIN",
    "RSUPER": "RWIN",
    # "": "APPS",
}

# "BACKSPACE": "BACKSPACE       DIK_BACK            /* backspace */
# "KP_MULTIPLY": "NUMPADSTAR      DIK_MULTIPLY        /* * on numeric keypad */
# "": "LALT            DIK_LMENU           /* left Alt */
# "": "CAPSLOCK        DIK_CAPITAL         /* CapsLock */
# "": "NUMPADMINUS     DIK_SUBTRACT        /* - on numeric keypad */
# "": "NUMPADPLUS      DIK_ADD             /* + on numeric keypad */
# "": "NUMPADPERIOD    DIK_DECIMAL         /* . on numeric keypad */
# "": "NUMPADSLASH     DIK_DIVIDE          /* / on numeric keypad */
# "": "RALT            DIK_RMENU           /* right Alt */
# "": "UPARROW         DIK_UP              /* UpArrow on arrow keypad */
# "": "PGUP            DIK_PRIOR           /* PgUp on arrow keypad */
# "": "LEFTARROW       DIK_LEFT            /* LeftArrow on arrow keypad */
# "": "RIGHTARROW      DIK_RIGHT           /* RightArrow on arrow keypad */
# "": "DOWNARROW       DIK_DOWN            /* DownArrow on arrow keypad */
# "": "PGDN            DIK_NEXT            /* PgDn on arrow keypad */



#!/usr/bin/python
"""Example bot for oyoyo that responds to !say"""

import logging
import re

from oyoyo.client import IRCClient
from oyoyo.cmdhandler import DefaultCommandHandler
from oyoyo import helpers


HOST = 'irc.freenode.net'
PORT = 6667
NICK = 'oyoyo-example'
CHANNEL = '#oyoyo-test'


class MyHandler(DefaultCommandHandler):
    def privmsg(self, nick, chan, msg):
        msg = msg.decode()
        match = re.match('\!say (.*)', msg)
        if match:
            to_say = match.group(1).strip()
            print('Saying, "%s"' % to_say)
            helpers.msg(self.client, chan, to_say)


def connect_cb(cli):
    helpers.join(cli, CHANNEL)


def main():
    logging.basicConfig(level=logging.DEBUG)

    cli = IRCClient(MyHandler, host=HOST, port=PORT, nick=NICK,
                    connect_cb=connect_cb)
    conn = cli.connect()

    while True:
        conn.next()      ## python 2
	# next(conn)     ## python 3


if __name__ == '__main__':
    main()



import fsui
from fsui.qt import QObject, QSignal
from .application import Application
from .button import Button, CustomButton, FlatButton
from .canvas import Canvas
from .color import Color
from .font import Font
from .image import Image
from .image_button import ImageButton
from .image_view import ImageView
from .label import Label, MultiLineLabel
from .layout import Column, Row
from .menu import Menu
from .painter import Painter
from .panel import Panel
from .text_area import TextArea
from .text_field import TextField
from .window import Window, TitlePanel, TitleSeparator


class Object(QObject):
    pass


# noinspection PyPep8Naming
def Signal():
    return QSignal()


def screen_size():
    return fsui.get_screen_size()



import fsui
from ...i18n import gettext
from .ImportDialog import ImportDialog


class ImportGroup(fsui.Group):

    AMIGA_FOREVER = 1

    def __init__(self, parent, import_type=0):
        fsui.Group.__init__(self, parent)
        self.type = import_type
        self.path = ""

        self.layout = fsui.VerticalLayout()
        if self.type == self.AMIGA_FOREVER:
            title = gettext("Import From Amiga Forever CD/DVD")
        else:
            title = gettext("Import Kickstarts and ROMs")
        label = fsui.HeadingLabel(self, title)
        self.layout.add(label, margin_bottom=10)

        icon_layout = fsui.HorizontalLayout()
        self.layout.add(icon_layout, fill=True)
        icon_layout.add_spacer(20)
        if self.type == self.AMIGA_FOREVER:
            image = fsui.Image("launcher:res/amiga_forever_group.png")
        else:
            image = fsui.Image("launcher:res/kickstart.png")
        self.image_view = fsui.ImageView(self, image)
        icon_layout.add(self.image_view, valign=0.0, margin_right=10)

        vert_layout = fsui.VerticalLayout()
        icon_layout.add(vert_layout, fill=True, expand=True)

        if self.type == self.AMIGA_FOREVER:
            text = gettext(
                "If you own Amiga Forever, select the drive/folder "
                'and click "{0}"'
            ).format(gettext("Import"))
        else:
            text = gettext(
                "Select a folder containing Amiga kickstart files "
                'and click "{0}"'
            ).format(gettext("Import"))
        label = fsui.Label(self, text)
        vert_layout.add(label, margin_bottom=10)

        hori_layout = fsui.HorizontalLayout()
        vert_layout.add(hori_layout, fill=True, margin=0)
        self.text_field = fsui.TextField(self, "", read_only=True)
        hori_layout.add(self.text_field, expand=True)
        self.browse_button = fsui.Button(self, gettext("Browse"))
        self.browse_button.activated.connect(self.on_browse)
        hori_layout.add(self.browse_button, margin_left=10)
        self.import_button = fsui.Button(self, gettext("Import"))
        self.import_button.activated.connect(self.on_import)
        self.import_button.disable()
        hori_layout.add(self.import_button, margin_left=10)

    def set_path(self, path):
        self.path = path
        self.text_field.set_text(path)
        self.import_button.enable()

    def on_browse(self):
        path = fsui.pick_directory(
            self.get_window(), gettext("Select Source Directory")
        )
        if path:
            self.set_path(path)

    def on_import(self):
        dialog = ImportDialog(self.get_window(), self.path, self.type)
        # dialog.show_modal()
        # dialog.destroy()
        dialog.show()
        # Signal.broadcast("scan_done")



from __future__ import print_function

import os
import StringIO

class ScanFile:
  """a file that is currently scanned"""

  def __init__(self, path, fobj, size, seekable=True, file_based=True):
    """create a scan file from a host file object"""
    if type(path) is list:
      self.paths = path
    else:
      self.paths = [path]
    self.fobj = fobj
    self.size = size
    self.seekable = seekable
    self.file_based = file_based

  def __str__(self):
    return "[%s:%d, seekable=%s, file_based=%s, fobj=%s]" % \
           (self.get_path(), self.size, self.seekable, self.file_based,
            self.fobj.__class__.__name__)

  def __repr__(self):
    return self.__str__()

  def is_seekable(self):
    return self.seekable

  def is_file_based(self):
    return self.file_based

  def get_path(self):
    return ";".join(self.paths)

  def get_local_path(self):
    return self.paths[-1]

  def get_basename(self):
    return os.path.basename(self.paths[-1])

  def get_fobj(self):
    return self.fobj

  def is_host_path(self):
    return len(self.paths) == 1

  def close(self):
    self.fobj.close()

  def create_sub_path(self, sub_path, fobj, size, seekable, file_based):
    paths = self.paths[:]
    paths.append(sub_path)
    return ScanFile(paths, fobj, size, seekable, file_based)

  def create_clone(self, seekable, file_based):
    src_fobj = self.fobj
    # create a temp file
    if file_based:
      fobj = tempfile.TemporaryFile()
      # copy original file
      blk_size = 4096
      while True:
        buf = src_fobj.read(blk_size)
        if len(buf) == 0:
          break
        fobj.write(buf)
    # create a string buffer
    else:
      data = src_fobj.read()
      fobj = StringIO.StringIO(data)
    # close old scan file
    src_fobj.close()
    # create promoted file
    return ScanFile(self.paths, fobj, self.size, seekable, file_based)



import sys
import subprocess
from distutils.spawn import find_executable

try:
    seven_zip_exe = find_executable(str("7z"))
except Exception:
    seven_zip_exe = None


class SevenZipFile:
    def __init__(self, path, mode="r"):
        if seven_zip_exe is None:
            raise Exception("no 7z executable found")

        assert mode == "r"
        self.path = path

        args = [seven_zip_exe, "l", "-slt", self.path]
        for i, arg in enumerate(args):
            args[i] = arg.encode(sys.getfilesystemencoding())
        print(args)
        p = subprocess.Popen(args, stdout=subprocess.PIPE)
        output = p.stdout.read()
        names = []
        last_name = None
        for line in output.split("\n"):
            name = line.strip().decode("UTF-8")
            print(name)
            if name.startswith("Path = "):
                last_name = name[7:]
                # names.append(name)
            elif name.startswith("Attributes = "):
                assert last_name
                if name.startswith("Attributes = D"):
                    names.append(last_name + "/")
                else:
                    names.append(last_name)
        # the first path is the name of the archive
        self.names = names[1:]
        # sort the name list so directory names are listed before contained
        # files
        self.names.sort()
        print(self.names)
        status = p.wait()
        if status != 0:
            raise Exception("7z status code not 0")

    def namelist(self):
        return self.names

    def read(self, name):
        args = [seven_zip_exe, "x", "-so", self.path, name]
        for i, arg in enumerate(args):
            args[i] = arg.encode(sys.getfilesystemencoding())
        print(args)
        p = subprocess.Popen(args, stdout=subprocess.PIPE)
        output = p.stdout.read()
        status = p.wait()
        if status != 0:
            raise Exception("7z status code not 0")
        return output

    def getinfo(self, name):
        if name in self.names:
            return {"name:": name}
        raise KeyError("{0} not found in archive".format(name))



import time
import traceback
from binascii import hexlify
from io import BytesIO as StringIO

import fsui
from fsbc.task import Task
from fsgs.archive import Archive
from fsgs.filedatabase import FileDatabase
from fsgs.amiga.rommanager import ROMManager
from fsgs.context import fsgs
from fsgs.ogd.client import OGDClient
from fsui.extra.iconheader import IconHeader
from launcher.res import gettext
from launcher.ui.widgets import CloseButton
from workspace.ui.theme import WorkspaceTheme


class LockerUploaderWindow(fsui.DialogWindow):
    @classmethod
    def open(cls, parent=None):
        return fsui.open_window_instance(cls, parent)

    def __init__(self, parent=None):
        title = gettext("OpenRetro Locker Uploader")
        super().__init__(parent, title)
        self.theme = WorkspaceTheme.instance()
        self.layout = fsui.VerticalLayout()
        self.set_icon(fsui.Icon("refresh", "pkg:workspace"))

        self.layout.min_width = 600
        self.layout.set_padding(20, 20, 20, 20)

        self.icon_header = IconHeader(
            self,
            fsui.Icon("refresh", "pkg:workspace"),
            gettext("OpenRetro Locker Uploader"),
            gettext("Upload recognized Amiga files to your OpenRetro locker"),
        )
        self.layout.add(self.icon_header, fill=True, margin_bottom=20)

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)
        self.created_label = fsui.Label(self, "")
        hori_layout.add(self.created_label, expand=True)

        self.upload_button = fsui.Button(self, gettext("Upload"))
        self.upload_button.activated.connect(self.on_upload_activated)
        hori_layout.add(self.upload_button, margin_left=10)

        self.stop_button = fsui.Button(self, gettext("Stop"))
        self.stop_button.activated.connect(self.on_stop_activated)
        self.stop_button.disable()
        hori_layout.add(self.stop_button, margin_left=10)

        if self.window.theme.has_close_buttons:
            self.close_button = CloseButton(self)
            hori_layout.add(self.close_button, margin_left=10)
        else:
            self.close_button = None

        self.set_size(self.layout.get_min_size())
        self.center_on_parent()
        self.task = None

    def __del__(self):
        print("LockerUploaderWindow.__del__")

    def on_close(self):
        if self.task is not None:
            self.task.stop()

    # def on_close_activated(self):
    #     self.close()

    def on_upload_activated(self):
        if self.close_button:
            self.close_button.disable()
        self.upload_button.disable()
        self.stop_button.enable()
        self.task = LockerUploaderTask()
        self.task.progressed.connect(self.on_progress)
        self.task.failed.connect(self.on_failure)
        self.task.succeeded.connect(self.on_success)
        self.task.stopped.connect(self.on_stopped)
        self.task.start()

    def on_stop_activated(self):
        self.task.stop()
        if self.close_button:
            self.stop_button.disable()

    def on_stopped(self):
        self.icon_header.subtitle_label.set_text(gettext("Stopped by user"))
        self.after_task_has_stopped()

    def on_success(self):
        self.after_task_has_stopped()

    def on_failure(self, message):
        fsui.show_error(message, parent=self.get_window())
        self.after_task_has_stopped()

    def after_task_has_stopped(self):
        if self.close_button:
            self.close_button.enable()
        self.upload_button.enable()
        self.stop_button.disable()

    def on_progress(self, message):
        if not isinstance(message, str):
            message = message[0]
        self.icon_header.subtitle_label.set_text(message)


def bytes_to_hex(sha1):
    return hexlify(sha1).decode("ASCII")


class LockerUploaderTask(Task):
    def __init__(self):
        Task.__init__(self, "Locker Uploader Task")
        self.client = OGDClient()

    def run(self):
        for i in range(16):
            self.upload_prefix(i)
        self.progressed(gettext("Locker upload task completed successfully"))

    def upload_prefix(self, prefix):
        self.stop_check()
        result = self.upload_check(prefix)
        print(len(result))
        for k in range(0, len(result), 20):
            self.stop_check()

            sha1 = result[k : k + 20]
            path = fsgs.file.find_by_sha1(bytes_to_hex(sha1))
            if not path:
                continue
            try:
                # this is done to properly handle encrypted ROMs
                archive = Archive(path)
                # FIXME: Use Archive.open to get support for filter functions
                # FIXME: Also use stream api, do not buffer entire file
                data = ROMManager.decrypt_archive_rom(archive, path)["data"]
            except Exception:
                traceback.print_exc()
                uri = "sha1://{0}".format(bytes_to_hex(sha1))
                print(uri)
                try:
                    input_stream = fsgs.file.open(uri)
                    data = input_stream.read()
                except Exception:
                    continue
                assert not input_stream.read()

            print("uploading file of size ", len(data))

            # self.progressed(gettext("Verifying {name}").format(
            #                 name=bytes_to_hex(sha1)))
            self.progressed(
                gettext("Uploading {name}").format(name=bytes_to_hex(sha1))
            )
            import hashlib

            new_hash = hashlib.sha1(data).hexdigest()
            print(new_hash, "vs", bytes_to_hex(sha1))
            if hashlib.sha1(data).hexdigest() != bytes_to_hex(sha1):
                print("hash mismatch, probably Cloanto ROM...")
                continue

            retry_seconds = 1
            while True:
                try:
                    self.client.post("/api/locker-upload-file", data=data)
                except OGDClient.NonRetryableHTTPError as e:
                    raise e
                except Exception:
                    traceback.print_exc()
                    self.progressed(
                        gettext("Re-trying in {0} seconds...").format(
                            retry_seconds
                        )
                    )
                    for _ in range(retry_seconds):
                        self.stop_check()
                        time.sleep(1.0)
                    retry_seconds = min(retry_seconds * 2, 60 * 10)
                else:
                    break

    def upload_check(self, prefix):
        self.progressed(
            gettext("Finding files eligible for OpenRetro Locker")
            + " ({:0.0%})".format((prefix / 16.0))
        )
        file_database = FileDatabase.instance()
        cursor = file_database.cursor()
        # FIXME: prefix
        p = "0123456789ABCDEF"[prefix]
        cursor.execute(
            "SELECT DISTINCT sha1 FROM file " "WHERE hex(sha1) LIKE ?",
            (p + "%",),
        )
        string_io = StringIO()
        for row in cursor:
            string_io.write(row[0])
        # print(prefix, len(string_io.getvalue()))
        self.stop_check()

        retry_seconds = 1
        while True:
            try:
                result = self.client.post(
                    "/api/locker-upload-check", data=string_io.getvalue()
                )
            except OGDClient.ForbiddenError:
                raise Task.Failure(
                    gettext(
                        "OpenRetro Locker is not enabled for your user. "
                        "It may be available only to a few select beta "
                        "users."
                    )
                )
            except OGDClient.NonRetryableHTTPError as e:
                raise e
            except Exception:
                traceback.print_exc()
                self.progressed(
                    gettext("Re-trying in {0} seconds...").format(
                        retry_seconds
                    )
                )
                for _ in range(retry_seconds):
                    self.stop_check()
                    time.sleep(1.0)
                retry_seconds = min(retry_seconds * 2, 60 * 10)
            else:
                return result


# application = SimpleApplication(LockerUploaderWindow)



import threading
import traceback
from .signal import Signal


local_tasks = threading.local()


class TaskFailure(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)
        self.message = message


class TaskStopped(Exception):
    def __init__(self):
        Exception.__init__(self)


class Task(object):

    Failure = TaskFailure
    Stopped = TaskStopped

    def __init__(self, name: str) -> None:
        self.__name = name
        self.stop_flag = False
        self.started = Signal()
        self.succeeded = Signal()
        self.progressed = Signal()
        self.failed = Signal()
        self.stopped = Signal()
        self.finished = Signal()

    def get_task_name(self):
        return self.__name

    def stop(self):
        self.stop_flag = True

    def stop_check(self):
        if self.stop_flag:
            print("raising TaskStopped for", self)
            raise TaskStopped()

    def start(self):
        threading.Thread(
            target=self.__run,
            name="TaskThread({0})".format(self.get_task_name()),
        ).start()

    def __run(self):
        local_tasks.task = self
        try:
            print(self, "starting")
            self.started()
            try:
                self.run()
            except TaskStopped:
                self.stopped()
            except TaskFailure as e:
                print(self, "failed", e.message)
                self.failed(e.message)
            except Exception as e:
                print(self, "failed", repr(str(e)))
                traceback.print_exc()
                self.failed(
                    "Task: {}\nError: {}\nMessage: {}\n\n"
                    "See the log file for more details.".format(
                        self.get_task_name(), type(e).__name__, str(e)
                    )
                )
            else:
                print(self, "succeeded")
                self.succeeded()
            self.finished.notify()
        finally:
            del local_tasks.task

    def set_progress(self, text):
        print(" --", text, "--")
        self.progressed.notify(text)

    def run(self):
        raise NotImplementedError("Task.run is not implemented")

    def __repr__(self):
        return '<Task "{0}">'.format(self.get_task_name())


class CurrentTaskProxy(object):
    @property
    def task(self):
        try:
            t = local_tasks.task
        except AttributeError:
            t = null_task
        return t

    @property
    def stop_flag(self):
        return self.task.stop_flag

    def set_progress(self, text):
        self.task.set_progress(text)

    def set_download_indicator(self, progress):
        pass

    def set_upload_indicator(self, progress):
        pass


null_task = Task("Null Task")
current_task = CurrentTaskProxy()



import fsui
from .TabPanel import TabPanel
from .TabButton import TabButton
from .Constants import Constants
from .skin import Skin


class WindowWithTabs(fsui.Window):
    def __init__(self, parent, title, border=True, menu=False):
        fsui.Window.__init__(
            self, parent, title, border=border, separator=False, menu=menu
        )
        Skin.set_background_color(self)
        self.toolbar = None
        self.tab_panel = TabPanel(self)
        self.layout = fsui.VerticalLayout()
        # if Skin.fws():
        #     self.title_bar = TitleBar(self)
        #     self.layout.add(self.title_bar, fill=True)
        if self.tab_panel:
            self.layout.add(self.tab_panel, fill=True)
        self.current_tab_group_id = 0
        self.tab_groups = [[]]

    # def __size_event(self, event):
    #     size = event.GetSize()
    #     print("WindowWithTabs size event, size =", size)
    #     event.Skip()

    def realize_tabs(self):
        if self.toolbar:
            self.toolbar.Realize()

    def new_tab_group(self):
        self.current_tab_group_id += 1
        self.tab_groups.append([])

    def set_content(self, content):
        self.layout.add(content, expand=True, fill=True)

    def select_tab(self, index, group):
        if self.toolbar:
            pass
        else:
            print("\n\n\nselect tab", index, group)
            # noinspection PyUnresolvedReferences
            self.tab_groups[group][index].select()
            # self.tab_groups[group].select_tab(index)

    def add_tab(self, function, icon, title="", tooltip=""):
        if not tooltip:
            tooltip = title
        button = TabButton(self.tab_panel, icon)
        button.set_tooltip(tooltip)
        button.group_id = self.current_tab_group_id
        button.on_select = function
        self.tab_panel.add(button)
        # noinspection PyTypeChecker
        self.tab_groups[self.current_tab_group_id].append(button)

    def add_tab_button(
        self,
        function,
        icon,
        title="",
        tooltip="",
        menu_function=None,
        left_padding=0,
        right_padding=0,
    ):
        if not tooltip:
            tooltip = title
        button = TabButton(
            self.tab_panel,
            icon,
            button_type=TabButton.TYPE_BUTTON,
            left_padding=left_padding,
            right_padding=right_padding,
        )
        button.set_tooltip(tooltip)
        button.group_id = self.current_tab_group_id
        menu_data = [None]
        if function:
            button.activated.connect(function)
        elif menu_function:

            def menu_wrapper():
                print("menu button click")
                print(menu_data[0] is not None and menu_data[0].is_open())
                if menu_data[0] is not None and menu_data[0].is_open():
                    menu_data[0].close()
                else:
                    menu_data[0] = menu_function()
                    button.check_hover()

            # def on_left_up():
            #     print("on left up")
            #     # menu_data[0].close()

            button.on_left_down = menu_wrapper
            button.on_left_dclick = menu_wrapper
            # button.on_left_up = on_left_up
        self.tab_panel.add(button)
        # noinspection PyTypeChecker
        self.tab_groups[self.current_tab_group_id].append(button)
        return button

    def add_tab_panel(self, class_, min_width=0, expand=1000000):
        # panel = class_(self.tab_panel, padding_bottom=2)
        panel = class_(self.tab_panel)
        panel.expandable = True
        panel.set_min_height(Constants.TAB_HEIGHT)
        if self.toolbar:
            panel.SetSize((min_width, 46))
            self.toolbar.AddControl(panel)
        else:
            self.tab_panel.add(panel, expand=expand)
        return panel

    def add_tab_separator(self):
        if self.toolbar:
            self.toolbar.AddSeparator()

    def add_tab_spacer(self, spacer=0, expand=False):
        self.tab_panel.add_spacer(spacer, expand=expand)



from fsgs.plugins.pluginexecutablefinder import PluginExecutableFinder
import os
import subprocess
import tempfile
import traceback

from fsbc.application import Application, app
from fsbc.system import windows, macosx
from fsgs.plugins.pluginmanager import Plugin


class FSUAE(object):
    @classmethod
    def start_with_config(cls, config, cwd=None):
        print("FSUAE.start_with_config:")
        tf = tempfile.NamedTemporaryFile(suffix=".fs-uae", delete=False)
        config_file = tf.name
        with tf:
            for line in config:
                print(line)
                tf.write(line.encode("UTF-8"))
                tf.write(b"\n")
        args = [config_file]
        return cls.start_with_args(args, cwd=cwd), config_file

    @classmethod
    def start_with_args(cls, args, cwd=None, **kwargs):
        print("FSUAE.start_with_args:", args)
        exe = PluginExecutableFinder().find_executable("fs-uae")
        if exe is None:
            raise Exception("Could not find fs-uae executable")
        print("current dir (cwd): ", os.getcwd())
        if cwd is not None:
            print("cwd override:", cwd)
        print("using fs-uae executable:", exe)
        args = [exe] + args

        # if "--always-on-top" in sys.argv:
        #     args += ["--always-on-top"]

        print(args)
        for name in ["SDL_VIDEODRIVER", "__GL_SYNC_TO_VBLANK"]:
            if name in os.environ:
                print("{0} was present in environment, removing!".format(name))
                del os.environ[name]
        env = os.environ.copy()
        print(repr(env))
        # env = None
        try:
            cls.center_window(args, env)
        except Exception:
            traceback.print_exc()
        cls.add_environment_from_settings(env)
        # env[str("SDL_VIDEO_WINDOW_POS")] = "0,0"
        # args += ["--fullscreen-mode", "desktop"]
        if windows:
            p = subprocess.Popen(
                args, cwd=cwd, env=env, close_fds=True, **kwargs
            )
        else:
            p = subprocess.Popen(args, cwd=cwd, env=env, **kwargs)
        return p

    @classmethod
    def add_environment_from_settings(cls, env):
        try:
            values = app.settings.values
        except AttributeError:
            values = {}
        for key, value in values.items():
            if not key.isupper():
                continue
            # Check if it looks like a valid environment variable
            for c in key:
                if c not in "ABCDEFGHIJKLMNOPQRSTUVWXYZ_0123456789":
                    break
            else:
                print("[ENV] {} = {}".format(key, value))
                env[key] = value

    @classmethod
    def center_window(cls, args, env):
        # FIXME: does not really belong here (dependency loop)
        from launcher.launcher_config import LauncherConfig
        from launcher.launcher_settings import LauncherSettings

        width = LauncherConfig.get("window_width") or LauncherSettings.get(
            "window_width"
        )
        height = LauncherConfig.get("window_height") or LauncherSettings.get(
            "window_height"
        )
        try:
            width = int(width)
        except:
            width = 960
        try:
            height = int(height)
        except:
            height = 540
        from launcher.ui.launcherwindow import LauncherWindow

        if LauncherWindow.current() is None:
            return

        main_w, main_h = LauncherWindow.current().get_size()
        main_x, main_y = LauncherWindow.current().get_position()

        x = main_x + (main_w - width) // 2
        y = main_y + (main_h - height) // 2

        # FIXME: re-implement without wx
        # if windows:
        #     import wx
        #    y += wx.SystemSettings_GetMetric(wx.SYS_CAPTION_Y)

        env[str("SDL_VIDEO_WINDOW_POS")] = str("{0},{1}".format(x, y))
        args.append("--window-x={0}".format(x))
        args.append("--window-y={0}".format(y))
        # print("window position", env["SDL_VIDEO_WINDOW_POS"])
        # os.environ["SDL_VIDEO_WINDOW_POS"] = "{0},{1}".format(x, y)



from __future__ import absolute_import
from __future__ import print_function

from .BlockScan import BlockScan
from amitools.fs.FSString import FSString
from amitools.fs.FileName import FileName
from amitools.fs.validate.Log import Log
import amitools.fs.DosType as DosType

class DirChainEntry:
  """entry of the hash chain"""
  def __init__(self, blk_info):
    self.blk_info = blk_info
    self.parent_ok = False
    self.fn_hash_ok = False
    self.valid = False
    self.end = False
    self.orphaned = False
    self.sub = None
    
  def __str__(self):
    l = []
    if self.parent_ok:
      l.append("parent_ok")
    if self.fn_hash_ok:
      l.append("fn_hash_ok")
    if self.valid:
      l.append("valid")
    if self.end:
      l.append("end")
    if self.orphaned:
      l.append("orphaned")
    return "[DCE @%d '%s': %s]" % \
      (self.blk_info.blk_num, self.blk_info.name, " ".join(l))


class DirChain:
  """representing a chain of the hashtable in a directory"""
  def __init__(self, hash_val):
    self.hash_val = hash_val
    self.chain = []
  
  def add(self, dce):
    self.chain.append(dce)
    
  def get_entries(self):
    return self.chain
  
  def __str__(self):
    return "{DirChain +%d: #%d}" % (self.hash_val, len(self.chain))
    

class DirInfo:
  """information structure on a directory"""
  def __init__(self, blk_info):
    self.blk_info = blk_info
    self.chains = {}
    self.children = []

  def add(self, dc):
    self.chains[dc.hash_val] = dc
  
  def add_child(self, c):
    self.children.append(c)
    
  def get(self, hash_val):
    if hash_val in self.chains:
      return self.chains[hash_val]
    else:
      return None
  
  def get_chains(self):
    return self.chains
  
  def __str__(self):
    bi = self.blk_info
    blk_num = bi.blk_num
    name = bi.name
    parent_blk = bi.parent_blk
    return "<DirInfo @%d '%s' #%d parent:%d child:#%d>" % (blk_num, name, len(self.chains), parent_blk, len(self.children))


class DirScan:
  """directory tree scanner"""
  
  def __init__(self, block_scan, log):
    self.log = log
    self.block_scan = block_scan
    self.root_di = None
    self.intl = DosType.is_intl(block_scan.dos_type)
    self.files = []
    self.dirs = []
  
  def scan_tree(self, root_blk_num, progress=None):
    """scan the root tree"""
    # get root block info
    root_bi = self.block_scan.get_block(root_blk_num)
    if root_bi == None:
      self.log.msg(Log.ERROR,"Root block not found?!",root_blk_num)      
      return None
    # do tree scan
    if progress != None:
      progress.begin("dir")
    self.root_di = self.scan_dir(root_bi, progress)
    if progress != None:
      progress.end()
    return self.root_di
  
  def scan_dir(self, dir_bi, progress):    
    """check a directory by scanning through the hash table entries and follow the chains
       Returns (all_chains_ok, dir_obj)
    """
    # create new dir info
    di = DirInfo(dir_bi)
    self.dirs.append(di)
    
    # run through hash_table of directory and build chains
    chains = {}
    hash_val = 0
    for blk_num in dir_bi.hash_table:
      if blk_num != 0:
        # build chain
        chain = DirChain(hash_val)
        self.build_chain(chain, dir_bi, blk_num, progress)
        di.add(chain)
      hash_val += 1
          
    return di
    
  def build_chain(self, chain, dir_blk_info, blk_num, progress):
    """build a block chain"""
    dir_blk_num = dir_blk_info.blk_num
    dir_name = dir_blk_info.name
    hash_val = chain.hash_val
    
    # make sure entry block is first used
    block_used = self.block_scan.is_block_available(blk_num)
    
    # get entry block
    blk_info = self.block_scan.read_block(blk_num)

    # create dir chain entry
    dce = DirChainEntry(blk_info)
    chain.add(dce)

    # account
    if progress != None:
      progress.add()

    # block already used?
    if block_used:
      self.log.msg(Log.ERROR, "dir block already used in chain #%d of dir '%s (%d)" % (hash_val, dir_name, dir_blk_num), blk_num)
      dce.end = True
      return

    # self reference?
    if blk_num == dir_blk_num:
      self.log.msg(Log.ERROR, "dir block in its own chain #%d of dir '%s' (%d)" % (hash_val, dir_name, dir_blk_num), blk_num)
      dce.end = True
      return
    
    # not a block in range
    if blk_info == None:
      self.log.msg(Log.ERROR, "out-of-range block terminates chain #%d of dir '%s' (%d)" % (hash_val, dir_name, dir_blk_num), blk_num)
      dce.end = True
      return
    
    # check type of entry block
    b_type = blk_info.blk_type
    if b_type not in (BlockScan.BT_DIR, BlockScan.BT_FILE_HDR):
      self.log.msg(Log.ERROR, "invalid block terminates chain #%d of dir '%s' (%d)" % (hash_val, dir_name, dir_blk_num), blk_num)
      dce.end = True
      return 
    
    # check referenceed block type in chain
    blk_type = blk_info.blk_type
    if blk_type in (BlockScan.BT_ROOT, BlockScan.BT_FILE_LIST, BlockScan.BT_FILE_DATA):
      self.log.msg(Log.ERROR, "invalid block type %d terminates chain #%d of dir '%s' (%d)" % (blk_type, hash_val, dir_name, dir_blk_num), blk_num)
      dce.end = True
      return

    # all following are ok
    dce.valid = True
     
    # check parent of block
    name = blk_info.name
    dce.parent_ok = (blk_info.parent_blk == dir_blk_num)
    if not dce.parent_ok:
      self.log.msg(Log.ERROR, "invalid parent in '%s' chain #%d of dir '%s' (%d)" % (name, hash_val, dir_name, dir_blk_num), blk_num)

    # check name hash
    fn = FileName(name, self.intl)
    fn_hash = fn.hash()
    dce.fn_hash_ok = (fn_hash == hash_val)
    if not dce.fn_hash_ok:
      self.log.msg(Log.ERROR, "invalid name hash in '%s' chain #%d of dir '%s' (%d)" % (name, hash_val, dir_name, dir_blk_num), blk_num)      

    # recurse into dir?
    if blk_type == BlockScan.BT_DIR:
      dce.sub = self.scan_dir(blk_info, progress)
    elif blk_type == BlockScan.BT_FILE_HDR:
      self.files.append(dce)

    # check next block in chain
    next_blk = blk_info.next_blk
    if next_blk != 0:
      self.build_chain(chain, dir_blk_info, next_blk, progress)
    else:
      dce.end = True
  
  def get_all_file_hdr_blk_infos(self):
    """return all file chain entries"""
    result = []
    for f in self.files:
      result.append(f.blk_info)    
    return result
  
  def get_all_dir_infos(self):
    """return all dir infos"""
    return self.dirs
  
  def dump(self):
    """dump whole dir info structure"""
    self.dump_dir_info(self.root_di, 0)
    
  def dump_dir_info(self, di, indent):
    """dump a single dir info structure and its sub dirs"""
    istr = "    " * indent
    print(istr, di)
    for hash_value in sorted(di.get_chains().keys()):
      dc = di.get(hash_value)
      print(istr, " ", dc)
      for dce in dc.get_entries():
        print(istr, "  ", dce)
        sub = dce.sub
        if sub != None and dce.blk_info.blk_type == BlockScan.BT_DIR:
          self.dump_dir_info(sub, indent+1)



import os
import shutil
import threading
import traceback
from fsbc.path import is_same_file
import fsui
from ...i18n import gettext
from fsgs.amiga.rommanager import ROMManager
from fsgs.FSGSDirectories import FSGSDirectories
from ...launcher_signal import LauncherSignal
from fsgs.filedatabase import FileDatabase


class ImportTask(threading.Thread):

    AMIGA_FOREVER = 1

    def __init__(self, path, import_type):
        threading.Thread.__init__(self, name="ImportTaskThread")
        self.path = path
        self.type = import_type
        self.done = False
        self.log_lines = []
        self.log_lock = threading.Lock()

    def run(self):
        if self.type:
            self.log(gettext("Amiga Forever import task started"))
        else:
            self.log(gettext("Kickstart import task started"))
        self.log("")
        print("ImportTask.run")
        try:
            count = self.run_task()
        except Exception as e:
            self.log("")
            self.log(repr(e))
            traceback.print_exc()
        else:
            if count == 0:
                self.log(gettext("Warning: No ROM files were found!"))
            else:
                self.log("")
                self.log(gettext("{} ROM files were imported").format(count))
        self.done = True
        print("ImportTask.run is done")
        self.log("")
        self.log(gettext("Import task is done"))

        def run_in_main():
            LauncherSignal.broadcast("scan_done")

        fsui.call_after(run_in_main)

    def get_new_log_lines(self, count):
        with self.log_lock:
            if len(self.log_lines) <= count:
                return []
            return self.log_lines[count:]

    def log(self, line):
        with self.log_lock:
            self.log_lines.append(line)

    def run_task(self):
        if self.type == 0:
            return self.import_roms()
        elif self.type == 1:
            return self.import_amiga_forever()

    def import_roms(self):
        return self.copy_roms(self.path, FSGSDirectories.get_kickstarts_dir())

    def import_amiga_forever(self):
        return self.copy_roms(
            os.path.join(self.path, "Amiga Files", "Shared", "rom"),
            FSGSDirectories.get_kickstarts_dir(),
        )

    def copy_file(self, src, dst):
        with self.log_lock:
            self.log_lines.append(gettext("Copy {0}\nto {1}").format(src, dst))
        if is_same_file(src, dst):
            self.log_lines.append(
                "- source and destination are the same, skipping..."
            )
            return
        if not os.path.exists(os.path.dirname(dst)):
            os.makedirs(os.path.dirname(dst))
        if os.path.exists(dst):
            # try to remove the file first, in case the file has read-only
            # permissions
            try:
                os.remove(dst)
            except Exception:
                pass
        shutil.copy(src, dst)

    def copy_roms(self, src, dst):
        count = 0
        if not os.path.isdir(src):
            self.log("{0} is not a directory".format(src))
            return count
        src_file = os.path.join(src, "rom.key")
        if os.path.exists(src_file):
            dst_file = os.path.join(dst, "rom.key")
            self.copy_file(src_file, dst_file)
        for file_name in os.listdir(src):
            name, ext = os.path.splitext(file_name)
            if ext not in [".rom"]:
                continue
            src_file = os.path.join(src, file_name)
            dst_file = os.path.join(dst, file_name)
            self.copy_file(src_file, dst_file)

            database = FileDatabase.get_instance()
            ROMManager.add_rom_to_database(dst_file, database, self.log)
            database.commit()
            count += 1
        return count



# #import pyapp
# from fengestad.game_center.database import Database
#
#
# class GameHistory:
#
#    @classmethod
#    def get_recent_ids(cls):
#        ids = []
#        k = 0
#        while True:
#            full_id = fs.config_string("RecentlyPlayed/%d" % k)
#            if not full_id:
#                break
#            ids.append(full_id.strip())
#            k += 1
#        return ids
#
#    @classmethod
#    def get_recently_played(cls, count):
#        game_list = []
#        for full_id in cls.get_recent_ids():
#            try:
#                name, platform = full_id.rsplit("(")
#                platform = platform[:-1].strip()
#                name = name.strip()
#                Database.cursor.execute("SELECT id, name, config, platform, "
#                        "imageratio, screenratio, year, publisher,
# developer, subtitle FROM Game "
#                        "WHERE name = ? AND platform = ?",
#                        (name, platform))
#                last_item = None
#                for row in Database.cursor:
#                    if last_item:
#                        if last_item["name"] == row[1] and \
#                                last_item["platform"] == row[3]:
#                            last_item["configurations"].append(row[2])
#                            continue
#                    item = {"id": row[0], "name": row[1], "configurations":
# [row[2]],
#                            "platform": row[3], "ratio": row[4],
# "screenratio": row[5],
#                            "year": row[6], "publisher": row[7],
# "developer": row[8],
#                            "subtitlepos": row[9] }
#                    game_list.append(item)
#                    last_item = item
#
#                #game_info = GameList.get().get_game(full_id)
#            except Exception:
#                pass
#        return game_list
#
#    def add_recently_played(self, game_info):
#        print("STUB: add_recently_played")
#        return
#        full_id = u"{name} ({platform})".format(**game_info)
#        ids = self.get_recent_ids()
#        try:
#            ids.remove(full_id)
#        except ValueError:
#            pass
#        ids.insert(0, full_id)
#        ids = ids[:15]
#        pyapp.user.ini.remove_section("RecentlyPlayed")
#        for i, full_id in enumerate(ids):
#            pyapp.user.ini.set("RecentlyPlayed/%d" % i, full_id)



import json

from fsgs.option import Option
from fsgs.platform import PlatformHandler
from fsgs.platforms.atari.hatariataridriver import (
    HatariAtariDriver,
    ATARI_MODEL_1040ST,
)
from fsgs.platforms.loader import SimpleLoader


class AtariSTPlatformHandler(PlatformHandler):
    PLATFORM_NAME = "Atari ST"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        return AtariSTLoader(fsgs)

    def get_runner(self, fsgs):
        return HatariAtariDriver(fsgs)


class AtariSTLoader(SimpleLoader):
    def load_files(self, values):
        file_list = json.loads(values["file_list"])
        # assert len(file_list) == 1
        if file_list[0]["name"].endswith(".st"):
            self.config["floppy_drive_0"] = "sha1://{0}/{1}".format(
                file_list[0]["sha1"], file_list[0]["name"]
            )
        if file_list[0]["name"].endswith(".stx"):
            self.config["floppy_drive_0"] = "sha1://{0}/{1}".format(
                file_list[0]["sha1"], file_list[0]["name"]
            )

    def load_extra(self, values):
        self.config[Option.ATARI_MODEL] = values["model"]
        if not self.config[Option.ATARI_MODEL]:
            self.config[Option.ATARI_MODEL] = ATARI_MODEL_1040ST



from fsgs.drivers.gamedriver import GameDriver
from fsgs.platform import PlatformHandler
from fsgs.amiga.valueconfigloader import ValueConfigLoader
from fsgs.amiga.fsuaeamigadriver import FSUAEAmigaDriver


class AmigaPlatformHandler(PlatformHandler):
    PLATFORM_NAME = "Amiga"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        loader = ValueConfigLoader()
        loader.uuid = fsgs.game.variant.uuid
        return loader

    def get_runner(self, fsgs) -> GameDriver:
        return FSUAEAmigaDriver(fsgs)



import os
import hashlib
from fsgs.archive import Archive


class ROMManager(object):
    @classmethod
    def add_rom_to_database(cls, path, database, log_function=None):
        if log_function is None:
            log_function = print
        try:
            archive = Archive(path)
            rom = ROMManager.decrypt_archive_rom(archive, path)
        except Exception:
            import traceback

            traceback.print_exc()
            return
        try:
            st = os.stat(path)
        except:
            log_function("Error stat-ing file {0}".format(repr(path)))
            return
        size = st.st_size
        mtime = int(st.st_mtime)
        log_function(
            'Adding ROM "{0}" to database (SHA-1: {1})'.format(
                path, rom["sha1"]
            )
        )
        database.delete_file(path=path)
        database.add_file(path=path, sha1=rom["sha1"], mtime=mtime, size=size)

    @classmethod
    def decrypt_archive_rom(cls, archive, path, file=None):
        print("decrypt_archive_rom", path)
        result = {}
        f = archive.open(path)
        data = f.read(len("AMIROMTYPE1"))
        out_data = b""
        sha1 = hashlib.sha1()
        if data != b"AMIROMTYPE1":
            # not encrypted, write raw data
            sha1.update(data)
            out_data += data
            data = f.read()
            sha1.update(data)
            out_data += data
            result["data"] = out_data
            result["sha1"] = sha1.hexdigest()
            cls.patch_rom(result)
            if file is not None:
                file.write(result["data"])
            return result

        key_path = archive.join(archive.dirname(path), "rom.key")
        key_archive = Archive(key_path)
        try:
            f2 = key_archive.open(key_path)
        except Exception:
            raise Exception("did not find rom.key to decrypt ROM with")
        print("using key file", key_path)
        # if not os.path.exists(key_file):
        #     raise Exception("did not find rom.key to decrypt ROM with")
        key_data = f2.read()
        f2.close()

        while True:
            data = f.read(len(key_data))
            if not data:
                break
            dec = []
            for i in range(len(data)):
                dec.append(data[i] ^ key_data[i])
            dec_data = bytes(dec)
            # if file is not None:
            #     file.write(dec_data)
            out_data += dec_data
            if sha1 is not None:
                sha1.update(dec_data)
        result["data"] = out_data
        result["sha1"] = sha1.hexdigest()
        cls.patch_rom(result)
        if file is not None:
            file.write(result["data"])
        return result

    @classmethod
    def patch_rom(cls, rom):
        src_data = rom["data"]
        dst_data = b""

        if rom["sha1"] == "c39bd9094d4e5f4e28c1411f3086950406062e87":
            dst_data += src_data[:413]
            dst_data += b"\x08"
            dst_data += src_data[414:176029]
            dst_data += b"\xb9"
            dst_data += src_data[176030:262121]
            dst_data += b"\x26"
            dst_data += src_data[262122:]
            rom["sha1"] = "891e9a547772fe0c6c19b610baf8bc4ea7fcb785"

        elif rom["sha1"] == "90933936cce43ca9bc6bf375662c076b27e3c458":
            # from Kickstart v1.3 rev 34.5 (1987)(Commodore)
            # (A500-A1000-A2000-CDTV)[o].rom
            dst_data += src_data[:262144]
            rom["sha1"] = "891e9a547772fe0c6c19b610baf8bc4ea7fcb785"

        elif rom["sha1"] == "c3c481160866e60d085e436a24db3617ff60b5f9":
            # from Kickstart v3.1 r40.68 (1993)(Commodore)(A4000)[h Cloanto]
            # to Kickstart v3.1 r40.68 (1993)(Commodore)(A4000)
            dst_data += src_data[:220]
            dst_data += b"\x74"
            dst_data += src_data[221:222]
            dst_data += b"\x7a"
            dst_data += src_data[223:326]
            dst_data += b"\x70"
            dst_data += src_data[327:434]
            dst_data += b"\x7c"
            dst_data += src_data[435:524264]
            dst_data += b"\x45"
            dst_data += src_data[524265:524266]
            dst_data += b"\x14"
            dst_data += src_data[524267:]
            rom["sha1"] = "5fe04842d04a489720f0f4bb0e46948199406f49"

        if dst_data:
            rom["data"] = dst_data
        sha1 = hashlib.sha1()
        sha1.update(rom["data"])
        assert sha1.hexdigest() == rom["sha1"]



import os

from fsgs.filedatabase import FileDatabase
from fsgs.util.gamenameutil import GameNameUtil
from .i18n import gettext


class ConfigurationScanner:
    def __init__(self, paths=None, on_status=None, stop_check=None):
        if paths is None:
            paths = []
        self.paths = paths
        self.on_status = on_status
        self._stop_check = stop_check
        self.scan_count = 0

    def stop_check(self):
        if self._stop_check:
            return self._stop_check()

    def set_status(self, title, status):
        if self.on_status:
            self.on_status((title, status))

    def scan_fs_uae_files(self, database):
        cursor = database.cursor()
        cursor.execute("SELECT id, path FROM game WHERE path IS NOT NULL")
        config_path_ids = {}
        for row in cursor:
            config_path_ids[database.decode_path(row[1])] = row[0]

        file_database = FileDatabase.get_instance()
        configurations = file_database.find_files(ext=".fs-uae")
        for c in configurations:
            if self.stop_check():
                break
            # name = os.path.basename(c["path"])
            # name = c["name"]
            # name = name[:-7]
            # search = name.lower()
            path = c["path"]
            name = os.path.basename(path)
            name, ext = os.path.splitext(name)
            sort_key = name.lower()
            # search = self.create_configuration_search(name)
            name = self.create_configuration_name(name)

            game_id = database.add_configuration(
                path=path, name=name, sort_key=sort_key
            )
            database.update_game_search_terms(
                game_id, self.create_search_terms(name)
            )

            print(config_path_ids)
            try:
                del config_path_ids[path]
            except KeyError:
                pass

                # FIXME: try splitting name into name, variant pair, and
                # actually add game_variant for config files as well, where
                # possible

        for id in config_path_ids.values():
            database.delete_game(id=id)

    def scan(self, database):
        self.set_status(
            gettext("Scanning configurations"), gettext("Please wait...")
        )

        self.set_status(
            gettext("Scanning configurations"),
            gettext("Scanning .fs-uae files..."),
        )
        self.scan_fs_uae_files(database)

        if self.stop_check():
            # aborted
            # database.rollback()
            return

            # database.remove_unscanned_configurations(self.scan_version)
            # print("remove unscanned games")
            # self.set_status(_("Scanning configurations"),
            #                 _("Purging old entries..."))
            # database.remove_unscanned_games_new(self.scan_version)
            # print("remove unscanned configurations")
            # database.remove_unscanned_configurations(self.scan_version)

            # self.set_status(_("Scanning configurations"), _("Committing
            # data..."))
            # database.commit()

    # @classmethod
    # def create_configuration_search(cls, name):
    #     return name.lower()

    @classmethod
    def create_search_terms(cls, name):
        search_terms = set()
        game_name = name.split("(", 1)[0]
        search_terms.update(GameNameUtil.extract_index_terms(game_name))
        return search_terms

    @classmethod
    def create_configuration_name(cls, name):
        if "(" in name:
            primary, secondary = name.split("(", 1)
            secondary = secondary.replace(", ", " \u00b7 ")
            # name = primary.rstrip() + " \u2013 " + secondary.lstrip()
            name = primary.rstrip() + "\n" + secondary.lstrip()
            if name[-1] == ")":
                name = name[:-1]
            name = name.replace(") (", " \u00b7 ")
            name = name.replace(")(", " \u00b7 ")
        return name



import fsui
from fsbc.util import unused
from launcher.ui.skin import Skin


class WizardHeader(fsui.Panel):
    def __init__(self, parent, icon, title):
        super().__init__(parent)
        self.layout = fsui.HorizontalLayout()
        # FIXME: From Theme
        self.set_background_color(fsui.Color(0xFF, 0xFF, 0xFF))

        self.title_label = fsui.HeadingLabel(self, title)
        font = fsui.Font("Roboto", 26)
        self.title_label.set_font(font)
        self.layout.add(self.title_label, expand=True, fill=True, margin=20)
        self.layout.add_spacer(20)
        self.layout.add(fsui.ImageView(self, icon.image(48)), margin_right=20)



from binascii import hexlify
import hashlib
import json
import os
import struct
from base64 import b64decode
from functools import lru_cache

import shutil

from fsbc.paths import Paths
from fsbc.system import System
from fsgs.archive import Archive
from fsgs.drivers.gamedriver import GameDriver, Emulator
from fsgs.input.enumeratehelper import EnumerateHelper
from fsgs.input.mapper import InputMapper
from fsgs.option import Option
from fsgs.saves import SaveHandler


class MednafenDriver(GameDriver):
    def __init__(self, fsgc, vanilla=False):
        super().__init__(fsgc)
        if vanilla:
            self.emulator = Emulator("mednafen")
        else:
            self.emulator = Emulator("mednafen-fs")
        self.save_handler = MednafenSaveHandler(
            self.fsgc, options=self.options
        )
        # self._game_files_added = False

    def prepare(self):
        if os.path.exists(os.path.join(self.home.path, ".mednafen")):
            shutil.rmtree(os.path.join(self.home.path, ".mednafen"))
        self.save_handler.prepare()
        # self.temp_home = self.create_temp_dir("mednafen-home")
        with open(self.mednafen_cfg_path(), "w", encoding="UTF-8") as f:
            self.mednafen_configure(f)
        # if not self._game_files_added:
        game_file = self.get_game_file()
        if game_file is not None:
            self.emulator.args.append(game_file)

    def finish(self):
        self.save_handler.finish()

    def set_mednafen_aspect(self, h, v):
        if self.stretching() == self.NO_STRETCHING:
            h, v = self.game_video_size()
        self.emulator.env["FSGS_ASPECT"] = "{}/{}".format(h, v)

    # FIXME: REPLACE BAD IMPLEMENTATION OF prepare cd images

    @staticmethod
    def expand_default_path(src, default_dir):
        if "://" in src:
            return src, None
        src = Paths.expand_path(src, default_dir)
        archive = Archive(src)
        return src, archive

    def prepare_mednafen_cd_images(self):
        # self._game_files_added = True

        temp_dir = self.temp_dir("media").path
        game_file = None
        # cdrom_drive_0 = self.config.get("cdrom_drive_0", "")
        # if cdrom_drive_0.startswith("game:"):
        if True:
            # scheme, dummy, game_uuid, name = cdrom_drive_0.split("/")
            # file_list = self.get_file_list_for_game_uuid(game_uuid)
            file_list = json.loads(self.options[Option.FILE_LIST])
            for file_item in file_list:
                src = self.fsgc.file.find_by_sha1(file_item["sha1"])

                src, archive = self.expand_default_path(src, None)
                dst_name = file_item["name"]
                # current_task.set_progress(dst_name)

                dst = os.path.join(temp_dir, dst_name)
                self.fsgc.file.copy_game_file(src, dst)

            # cue_sheets = self.get_cue_sheets_for_game_uuid(game_uuid)
            cue_sheets = json.loads(self.options[Option.CUE_SHEETS])
            for i, cue_sheet in enumerate(cue_sheets):
                # FIXME: Try to get this to work with the PyCharm type checker
                path = os.path.join(temp_dir, cue_sheet["name"])
                if i == 0:
                    game_file = path
                # noinspection PyTypeChecker
                with open(path, "wb") as f:
                    # noinspection PyTypeChecker
                    f.write(cue_sheet["data"].encode("UTF-8"))

            if self.options[Option.SBI_DATA]:
                sbi_data = json.loads(self.options[Option.SBI_DATA])
                for i, sbi_file in enumerate(sbi_data):
                    path = os.path.join(temp_dir, sbi_file["name"])
                    with open(path, "wb") as f:
                        f.write(b64decode(sbi_file["base64"]))

        self.emulator.args.append(game_file)

    def prepare_mednafen_bios(self, known_file, name):
        bios_path = os.path.join(self.home.path, ".mednafen", name)
        if not os.path.exists(os.path.dirname(bios_path)):
            os.makedirs(os.path.dirname(bios_path))
        src = self.fsgc.file.find_by_sha1(known_file.sha1)
        if not src:
            raise Exception(
                "Could not find {} (SHA-1: {}".format(
                    known_file.name, known_file.sha1
                )
            )
        self.fsgc.file.copy_game_file(src, bios_path)

    def get_supported_filters(self):
        supported = [
            {"name": "2x", "special": "nn2x"},
            {"name": "none", "special": "none"},
            {"name": "scale2x", "special": "scale2x"},
            {"name": "hq2x", "special": "hq2x"},
        ]
        return supported

    def mednafen_input_mapping(self, port):
        raise NotImplementedError()

    def mednafen_system_prefix(self):
        raise NotImplementedError()

    def game_video_par(self):
        return 1.0

    def game_video_size(self):
        raise NotImplementedError("game_video_size must be implemented")

    def mednafen_rom_extensions(self):
        return []

    def mednafen_scanlines_setting(self):
        return None

    def mednafen_special_filter(self):
        return None

    def mednafen_viewport(self):
        viewport = self.options["viewport"]
        if viewport:
            src, dst = viewport.split("=")
            src = src.strip()
            dst = dst.strip()
            src_x, src_y, src_w, src_h = [int(v) for v in src.split(" ")]
            dst_x, dst_y, dst_w, dst_h = [int(v) for v in dst.split(" ")]
            return (src_x, src_y, src_w, src_h), (dst_x, dst_y, dst_w, dst_h)
        return None, None

    def mednafen_configure(self, f):
        pfx = self.mednafen_system_prefix()
        self.configure_audio(f)
        self.configure_input(f)
        self.configure_video(f)

        cheats_file_name = self.mednafen_system_prefix() + ".cht"
        cheats_file_path = self.cheats_file(cheats_file_name)
        if cheats_file_path:
            self.emulator.args.extend(["-cheats", "0"])
            cheats_dir = self.temp_dir("cheats").path
            shutil.copy(
                cheats_file_path, os.path.join(cheats_dir, cheats_file_name)
            )
            # self.emulator.args.extend(["-filesys.path_cheat", cheats_dir])
            self.emulator.args.extend(["-path_cheat", cheats_dir])

        print("\n" + "-" * 79 + "\n" + "CONFIGURE DIRS")

        self.emulator.args.extend(
            ["-path_sav", self.save_handler.emulator_save_dir()]
        )
        self.emulator.args.extend(
            ["-path_state", self.save_handler.emulator_state_dir()]
        )

        # self.emulator.args.extend(["-filesys.fname_state", "%M%X"])
        # self.emulator.args.extend(["-filesys.fname_sav", "%M%x"])

        self.emulator.args.extend(["-filesys.fname_state", "%f.%X"])
        self.emulator.args.extend(["-filesys.fname_sav", "%f.%x"])

        # docdir = pyapp.user.documents_dir()
        self.doc_dir = self.temp_dir("mednafen-docs")
        self.emulator.args.extend(
            [
                "-path_movie",
                self.doc_dir.path,
                # "-path_cheat", self.doc_dir.path,
                "-path_palette",
                self.home.path,
            ]
        )

        self.emulator.args.extend(["-path_snap", self.screenshots_dir()])
        self.emulator.args.extend(
            [
                "-filesys.fname_snap",
                "{0}-%p.%x".format(self.screenshots_name()),
            ]
        )

        self.mednafen_post_configure()

    def configure_audio(self, _):
        # pfx = self.mednafen_system_prefix()
        audio_driver = self.options[Option.MEDNAFEN_AUDIO_DRIVER]
        if audio_driver in ["", "auto"]:
            if System.windows:
                pass
            elif System.macos:
                pass
            else:
                audio_driver = "sdl"
        if audio_driver == "sdl":
            # Mednafen does not support PulseAudio directly, but using the
            # sdl driver will "often" result in PulseAudio being used
            # indirectly.
            self.emulator.args.extend(["-sound.driver", "sdl"])
        elif audio_driver == "alsa":
            self.emulator.args.extend(
                ["-sound.device", "sexyal-literal-default"]
            )
        else:
            # Use Mednafen default selection
            # self.emulator.args.extend(["-sound.device", "default"])
            pass

        default_buffer_size = 40
        buffer_size = default_buffer_size
        if self.options[Option.MEDNAFEN_AUDIO_BUFFER]:
            try:
                buffer_size = int(self.options[Option.MEDNAFEN_AUDIO_BUFFER])
            except ValueError:
                print("WARNING: Invalid Mednafen audio buffer size specified")
            else:
                if buffer_size < 0 or buffer_size > 1000:
                    print("WARNING: Mednafen audio buffer size out of range")
                    buffer_size = default_buffer_size
        self.emulator.args.extend(["-sound.buffer_time", str(buffer_size)])

    def configure_input(self, f):
        print("\n" + "-" * 79 + "\n" + "CONFIGURE PORTS")
        for i, port in enumerate(self.ports):
            input_mapping = self.mednafen_input_mapping(i)
            mapper = MednafenInputMapper(port, input_mapping)
            keys = {}
            for key, value in mapper.items():
                keys.setdefault(key, []).append(value)
            for key, values in keys.items():
                print(repr(key), repr(values))
                f.write(
                    "{key} {value}\n".format(
                        key=key, value=" || ".join(values)
                    )
                )

    def configure_video(self, f):
        pfx = self.mednafen_system_prefix()
        self.emulator.args.extend(["-video.driver", "opengl"])

        screen_w, screen_h = self.screen_size()
        # screen_a = screen_w / screen_h
        border_w, border_h = 0, 0

        # if self.border() == self.NO_BORDER:
        #     pass
        # else:
        #     border_w = screen_h * 32 / 1080
        #     border_h = screen_h * 32 / 1080

        dest_w, dest_h = screen_w - border_w, screen_h - border_h
        if not self.use_fullscreen():
            dest_w, dest_h = 960, 540
        # dest_a = dest_w /dest_h

        viewport_src, viewport = self.mednafen_viewport()
        if viewport is None:
            game_w, game_h = self.game_video_size()
        else:
            game_w, game_h = viewport[2], viewport[3]
        assert game_w
        assert game_h
        # game_a = game_w / game_h
        pixel_a = self.game_video_par()

        if self.scaling() == self.MAX_SCALING:
            if self.stretching() == self.STRETCH_FILL_SCREEN:
                scale_y = dest_h / game_h
                scale_x = dest_w / game_w
            elif self.stretching() == self.STRETCH_ASPECT:
                # FIXME: HANDLE TALL RESOLUTIONS
                scale_y = dest_h / game_h
                # scale_x = dest_w / (game_w * pixel_a)
                scale_x = scale_y * pixel_a
            else:  # Square pixels
                scale_y = dest_h / game_h
                scale_x = scale_y
        elif self.scaling() == self.INTEGER_SCALING:
            if self.stretching() == self.STRETCH_FILL_SCREEN:
                scale_y = dest_h // game_h
                scale_x = dest_w // game_w
            elif self.stretching() == self.STRETCH_ASPECT:
                # FIXME: HANDLE TALL RESOLUTIONS
                scale_y = dest_h // game_h
                scale_x = scale_y * pixel_a
            else:  # Square pixels
                scale_y = dest_h // game_h
                scale_x = scale_y
        else:  # NO SCALING
            scale_x = 1
            scale_y = 1

        if self.use_doubling():
            if scale_x == 1 and scale_y == 1:
                scale_x = 2
                scale_y = 2

        # self.emulator.args.extend(["-{}.special".format(pfx), "nn2x"])

        if self.smoothing() == self.SMOOTHING:
            videoip = "1"
        elif self.smoothing() == self.NON_INTEGER_SMOOTHING:
            if abs(scale_x - scale_x // 1) < 0.01:
                videoip = "y"
            else:
                videoip = "1"
            if abs(scale_y - scale_y // 1) < 0.01:
                if videoip == "y":
                    videoip = "0"
                else:
                    videoip = "x"
        else:
            videoip = "0"
        self.emulator.args.extend(["-{}.videoip".format(pfx), videoip])

        self.emulator.args.extend(["-{}.xscale".format(pfx), str(scale_x)])
        self.emulator.args.extend(["-{}.yscale".format(pfx), str(scale_y)])
        self.emulator.args.extend(["-{}.xscalefs".format(pfx), str(scale_x)])
        self.emulator.args.extend(["-{}.yscalefs".format(pfx), str(scale_y)])
        self.emulator.args.extend(["-{}.stretch".format(pfx), "0"])

        # Only enable v-sync if game refresh matches screen refresh.
        if self.configure_vsync():
            self.emulator.args.extend(["-video.glvsync", "1"])
        else:
            self.emulator.args.extend(["-video.glvsync", "0"])

        # Specify fullscreen size and conditionally enable fullscreen mode.
        self.emulator.args.extend(["-{}.xres".format(pfx), str(screen_w)])
        self.emulator.args.extend(["-{}.yres".format(pfx), str(screen_h)])
        if self.use_fullscreen():
            self.emulator.args.extend(["-fs", "1"])
        else:
            self.emulator.args.extend(["-fs", "0"])

        if self.effect() == self.CRT_EFFECT:
            self.emulator.args.extend(["-{}.shader".format(pfx), "goat"])
            self.emulator.args.extend(
                ["-{}.shader.goat.slen".format(pfx), "1"]
            )
            special = "none"
            video_scale = 2
            min_video_scale = 2
        elif self.effect() == self.DOUBLE_EFFECT:
            special = "nn2x"
            video_scale = 2
            min_video_scale = 1
        elif self.effect() == self.HQ2X_EFFECT:
            special = "hq2x"
            video_scale = 2
            min_video_scale = 2
        elif self.effect() == self.SCALE2X_EFFECT:
            special = "scale2x"
            video_scale = 2
            min_video_scale = 2
        else:
            special = "none"
            video_scale = 1
            min_video_scale = 1

        if self.scaling() == self.MAX_SCALING:
            window_w, window_h = 960, 540
        elif self.scaling() == self.INTEGER_SCALING:
            window_w = game_w * min_video_scale
            window_h = game_h * min_video_scale
            s = min_video_scale + 1
            print(window_w, window_h)
            while game_w * s <= 900 and game_h * s <= 700:
                window_w = game_w * s
                window_h = game_h * s
                s += 1
        else:
            window_w = game_w * video_scale
            window_h = game_h * video_scale
        self.emulator.env["FSGS_WINDOW_SIZE"] = "{},{}".format(
            window_w, window_h
        )

        deinterlacer = self.options[Option.MEDNAFEN_DEINTERLACER]
        if deinterlacer:
            self.emulator.args.extend(["-video.deinterlacer", deinterlacer])

        temporal_blur = self.options[Option.MEDNAFEN_TEMPORAL_BLUR] == "1"
        if temporal_blur:
            self.emulator.args.extend(["-{}.tblur".format(pfx), "1"])

        self.emulator.args.extend(["-{}.special".format(pfx), special])
        self.emulator.args.extend(self.mednafen_extra_graphics_options())

    def set_mednafen_input_order(self):
        if System.windows:
            self.input_device_order = "DINPUT8"
        self.input_mapping_multiple = False

    def mednafen_extra_graphics_options(self):
        return []

    def mednafen_post_configure(self):
        # can be overridden by subclasses
        pass

    def mednafen_cfg_path(self):
        if not os.path.exists(os.path.join(self.home.path, ".mednafen")):
            os.makedirs(os.path.join(self.home.path, ".mednafen"))
        # return os.path.join(self.home.path, ".mednafen", "mednafen-09x.cfg")
        return os.path.join(self.home.path, ".mednafen", "mednafen.cfg")

    def is_pal(self):
        # return self.config.get("ntsc_mode") != "1"
        # return False
        refresh_rate = self.get_game_refresh_rate()
        if refresh_rate:
            return int(round(refresh_rate)) == 50


class MednafenInputMapper(InputMapper):
    def __init__(self, port, mapping):
        InputMapper.__init__(self, port, mapping)
        helper = EnumerateHelper()
        helper.init()
        seen_ids = set()
        self.id_map = {}
        for device in helper.devices:
            uid = self.calculate_unique_id(device)
            while uid in seen_ids:
                uid += 1
            seen_ids.add(uid)
            self.id_map[device.id] = uid
        print("MednafenInputMapper device map")
        for id, uid in self.id_map.items():
            print(uid, id)

    def axis(self, axis, positive):
        if positive:
            sign = "+"
        else:
            sign = "-"
        joystick_id = self.unique_id(self.device, self.device.id)
        return "joystick 0x{:032x} abs_{}{}".format(joystick_id, axis, sign)

    def hat(self, hat, direction):
        offset = {"left": 3, "right": 1, "up": 0, "down": 2}[direction]
        joystick_id = self.unique_id(self.device, self.device.id)
        # Hats after buttons, order: up, right, down, left
        return "joystick 0x{:032x} button_{}".format(
            joystick_id, self.device.buttons + hat * 4 + offset
        )

    def button(self, button):
        joystick_id = self.unique_id(self.device, self.device.id)
        return "joystick 0x{:032x} button_{}".format(joystick_id, button)

    def key(self, key):
        # FIXME: Need other key codes on Windows ... ?
        # print(key)
        return "keyboard 0x0 {}".format(key.sdl2_scan_code)

    @lru_cache()
    def unique_id(self, device, _):
        try:
            return self.id_map[device.id]
        except KeyError:
            print("id_map:", self.id_map)
            raise

    @lru_cache()
    def calculate_unique_id(self, device, version=2):
        """Implements the joystick ID algorithm in mednafen.
        Was src/drivers/joystick.cpp:GetJoystickUniqueID
        Now src/drivers/Joystick.cpp:Calc09xID.
        """
        print("get_unique_id for", device.id)
        if version == 2:
            print(device)
            m = hashlib.md5()
            print("--------------{}----------------".format(device.sdl_name))
            m.update((device.sdl_name).encode("UTF-8"))
            # print(m.hexdigest()[:16], device.axes, device.buttons, device.hats, device.balls)

            buffer = struct.pack(
                ">HHHH", device.axes, device.buttons, device.hats, device.balls
            )
            return int(
                "{}{}".format(
                    hexlify(m.digest()[:8]).decode("ASCII"),
                    hexlify(buffer).decode("ASCII"),
                ),
                16,
            )
            # return "0x{}{:02x}{:02x}{:02x}{:02x}".format(
            #     m.hexdigest()[:16], device.axes, device.buttons, device.hats, device.balls)
        else:
            m = hashlib.md5()
            print(device.axes, device.balls, device.hats, device.buttons)
            # noinspection SpellCheckingInspection
            buffer = struct.pack(
                "iiii", device.axes, device.balls, device.hats, device.buttons
            )
            m.update(buffer)
            digest = m.digest()
            ret = 0
            for x in range(16):
                # ret ^= ord(digest[x]) << ((x & 7) * 8)
                ret ^= digest[x] << ((x & 7) * 8)
            # return "{:x}".format(ret)
            return ret


class MednafenSaveHandler(SaveHandler):
    def __init__(self, fsgc, options):
        super().__init__(fsgc, options, emulator="Mednafen")
        self._srm_alias = ""

    def prepare(self):
        self.copy_to_srm_alias()
        super().prepare()

    def finish(self):
        self.move_from_srm_alias()
        super().finish()

    def set_srm_alias(self, alias):
        assert alias.startswith(".")
        self._srm_alias = alias

    def copy_to_srm_alias(self):
        if not self._srm_alias:
            return
        save_dir = self.save_dir()
        if not os.path.exists(save_dir):
            return
        for full_name in os.listdir(save_dir):
            src = os.path.join(save_dir, full_name)
            if not os.path.isfile(src):
                continue
            name, ext = os.path.splitext(full_name)
            if ext not in [".srm"]:
                continue
            dst = os.path.join(save_dir, name + self._srm_alias)
            print("MednafenSaveHandler:", src, "->", dst)
            shutil.copy(src, dst)

    def move_from_srm_alias(self):
        if not self._srm_alias:
            return
        save_dir = self.save_dir()
        for full_name in os.listdir(save_dir):
            src = os.path.join(save_dir, full_name)
            if not os.path.isfile(src):
                continue
            name, ext = os.path.splitext(full_name)
            if ext != self._srm_alias:
                continue
            dst = os.path.join(save_dir, name + ".srm")
            print("MednafenSaveHandler:", dst, "<-", src)
            shutil.copy(src, dst)
            os.remove(src)



import os

from fsbc.system import windows, macosx


class GameCenter(object):
    @classmethod
    def register_user_activity(cls):
        if windows:
            pass
        elif macosx:
            pass
        else:
            path = "/tmp/fs-game-center-activity"
            if not os.path.exists(path):
                try:
                    with open(path, "wb") as _:
                        pass
                except Exception:
                    pass
                try:
                    os.chmod(path, 0o666)
                except Exception:
                    pass
            os.utime("/tmp/fs-game-center-activity", None)



import sys
import fsboot
import workspace
import workspace.ui


class RunApplication(workspace.ui.Application):
    """An application for browsing the virtual file hierarchy."""

    def __init__(self):
        super().__init__("Run")

    def init(self, args):
        pass


def app_main():
    fsboot.set("fws", "1")
    # Calculate base dir before we start messing with sys.argv[0]
    fsboot.base_dir()

    app = RunApplication()
    import workspace.shell

    if len(sys.argv) == 1:
        workspace.shell.workspace_open("SYS:System/FileManager", ["VOLUMES:"])
    else:
        workspace.shell.workspace_open(sys.argv[1], sys.argv[2:])
    app.run()

    from fsbc.application import Application

    Application.wait_for_threads()



import weakref
from fsui.qt import Qt, QSize, QAbstractListModel, QListView, QBrush, Signal
from .widget_mixin import WidgetMixin


class Model(QAbstractListModel):
    def __init__(self, parent):
        QAbstractListModel.__init__(self, parent)
        self.parent = weakref.ref(parent)
        self.count = 0

    # def set_item_count(self, count):
    #     self.count = count

    def rowCount(self, parent):
        # print("returning count", self.count, "for parent", parent)
        # return self.count
        return self.parent().get_item_count()

    def data(self, index, role):
        row = index.row()
        # print("data for", index, "role", role)
        if role == Qt.SizeHintRole:
            height = self.parent()._row_height
            return QSize(height, height)
        elif role == Qt.DecorationRole:
            icon = self.parent().get_item_icon(row)
            if icon:
                return icon.qpixmap
        elif role == Qt.DisplayRole:
            return self.parent().get_item_text(row)
        elif role == Qt.ForegroundRole:
            color = self.parent().get_item_text_color(row)
            if color is not None:
                return QBrush(color)
        # return QVariant()


class VerticalItemView(QListView, WidgetMixin):

    item_selected = Signal(int)
    item_activated = Signal(int)

    def __init__(self, parent, border=True):
        QListView.__init__(self, parent.get_container())
        # Widget.__init__(self, parent)
        self.init_widget(parent)
        self.viewport().installEventFilter(self.get_window())
        self.verticalScrollBar().installEventFilter(self.get_window())
        if not border:
            self.setFrameStyle(0)

        # self.setSelectionModel()
        # self.model = QStandardItemModel(self)
        self.model = Model(self)
        self.setModel(self.model)
        # self.itemSelectionChanged.connect(self.__selection_changed)
        selection_model = self.selectionModel()
        print("VerticalItemView selectionModel = ", selection_model)
        selection_model.selectionChanged.connect(self.__selection_changed)
        self.doubleClicked.connect(self.__double_clicked)
        # self.returnPressed.connect(self.__double_clicked)
        # self.activated.connect(self.__double_clicked)
        self._row_height = 26

    def set_row_height(self, height):
        self._row_height = height

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_Return:
            self.__double_clicked()
        else:
            QListView.keyPressEvent(self, event)

    def __double_clicked(self):
        index = self.get_index()
        if index is not None:
            self.on_activate_item(index)
            self.item_activated.emit(index)

    def __selection_changed(self):
        index = self.get_index()
        self.on_select_item(index)
        self.item_selected.emit(index)

    def on_activate_item(self, index):
        pass

    def get_item_icon(self, index):
        return None

    def get_item_text(self, index):
        return ""

    def get_item_text_color(self, index):
        return None

    # def set_item_count(self, count):
    #     #self.model.rowCoun
    #     self.model.set_item_count(count)
    #     #self.update()
    #     #self.invalidate()
    #     self.dataChanged(self.model.createIndex(0, 0),
    #             self.model.createIndex(count, 0))

    def set_default_icon(self, image):
        pass

    # def set_items(self, items):
    #    #print("set_items", items)
    #    self.model.clear()
    #    for label in items:
    #        item = QStandardItem(label)
    #        self.model.appendRow(item)

    def get_index(self):
        indices = self.selectionModel().selectedIndexes()
        if len(indices) == 0:
            return None
        return indices[0].row()

    def set_index(self, index):
        if index is None:
            index = -1
        # print(self.rootIndex)
        # idx = QModelIndex.createIndex(index)
        idx = self.model.index(index, 0)
        self.scrollTo(idx)
        self.setCurrentIndex(idx)

    def select_item(self, index):
        self.set_index(index)

    def on_select_item(self, index):
        pass

    def update(self):
        # self.model.rowCoun
        count = self.get_item_count()
        # self.model.set_item_count(count)
        # self.update()
        # self.invalidate()
        self.dataChanged(
            self.model.createIndex(0, 0), self.model.createIndex(count, 0)
        )



# FSGS - Common functionality for FS Game System.
# Copyright (C) 2013-2016  Frode Solheim <frode@openretro.org>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
"""
FSGS Game Driver for DOS.

FIXME: Save data

IMGMOUNT in autoexec can refer to paths beginning with $DRIVES, which will be
replaced by this driver when creating dosbox config.

"""
import hashlib
import json
import os
import shutil

from fsbc.path import str_path
from fsbc.system import System
from fsgs.FSGSDirectories import FSGSDirectories
from fsgs.drivers.gamedriver import GameDriver
from fsgs.option import Option
from fsgs.saves import SaveHandler


class DosBoxDosDriver(GameDriver):
    def __init__(self, fsgs):
        super().__init__(fsgs)
        if self.options[Option.DOS_EMULATOR] == "dosbox":
            self.emulator.name = "dosbox"
        elif self.options[Option.DOS_EMULATOR] == "dosbox-svn":
            self.emulator.name = "dosbox-svn"
        else:
            self.emulator.name = "dosbox-fs"

        self.save_handler = SaveHandler(self.fsgc, options=self.options)

        # Right now G-SYNC only works fine with 60Hz modes. There's stuttering
        # when DOSBox outputs 70 fps. Edit: The problem is that G-SYNC is
        # capped to monitors selected refresh rate.
        if self.screen_refresh_rate() < 70.0:
            print("[VIDEO] Screen refresh rate is < 70.0, disabling G-SYNC")
            self.set_allow_gsync(False)

        # self.ultrasnd_drive = None
        self.drives_dir = self.temp_dir("drives")
        self.drives = []

    def __del__(self):
        print("DosBoxDosDriver.__del__")

    def prepare(self):
        config_file = self.temp_file("dosbox.cfg").path
        self.prepare_media()
        with open(config_file, "w", encoding="UTF-8") as f:
            self.configure(f)
        self.emulator.args.extend(["-conf", config_file])

        # FIXME: Move to game driver?
        if self.options[Option.VIEWPORT]:
            self.emulator.env["FSGS_VIEWPORT"] = self.options[Option.VIEWPORT]

        # Must run after game hard drives have been unpacked
        for drive, drive_path in self.drives:
            assert drive in "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            # print(drive)
            # print(drive_path)
            self.save_handler.register_changes(
                drive_path, os.path.join(self.save_handler.save_dir(), drive)
            )
        self.save_handler.prepare()

    def finish(self):
        self.save_handler.finish()

    def prepare_media(self):
        file_list = json.loads(self.options[Option.FILE_LIST])
        self.unpack_game_hard_drives(file_list)

        if self.options["cue_sheets"]:
            cue_sheets = json.loads(self.options["cue_sheets"])
            for cue_sheet in cue_sheets:
                with open(
                    os.path.join(self.drives_dir.path, cue_sheet["name"]), "wb"
                ) as f:
                    f.write(cue_sheet["data"].encode("UTF-8"))

    def unpack_game_hard_drives(self, file_list):
        drives_added = set()
        dir_path = self.drives_dir.path
        for file_entry in file_list:
            # if self.stop_flag:
            #     return

            name = file_entry["name"]
            name = name.upper()

            drives_added.add(name[0])

            # Extract relative path and convert each path component
            # to host file name (where needed).

            rel_path = name
            print("rel_path", rel_path)
            rel_parts = rel_path.split("/")
            # for i, part in enumerate(rel_parts):
            #     # part can be blank if rel_parts is a directory
            #     # (ending with /)
            #     if part:
            #         rel_parts[i] = amiga_filename_to_host_filename(part)
            rel_path = "/".join(rel_parts)

            dst_file = os.path.join(dir_path, rel_path)
            print(repr(dst_file))
            if name.endswith("/"):
                os.makedirs(str_path(dst_file))
                continue
            sha1 = file_entry["sha1"]
            src_file = self.fsgc.file.find_by_sha1(sha1)
            if not os.path.exists(os.path.dirname(dst_file)):
                os.makedirs(os.path.dirname(dst_file))
            stream = self.fsgc.file.open(src_file)
            # archive = Archive(src_file)
            # f = archive.open(src_file)
            data = stream.read()
            assert hashlib.sha1(data).hexdigest() == sha1

            with open(dst_file, "wb") as out_file:
                out_file.write(data)
            if dst_file.endswith(".CUE"):
                with open(dst_file, "r", encoding="ISO-8859-1") as f:
                    data = f.read()
                data = data.upper()
                with open(dst_file, "w", encoding="ISO-8859-1") as f:
                    f.write(data)

        for drive in sorted(drives_added):
            self.drives.append((drive, os.path.join(dir_path, drive)))

    def configure(self, f):
        f.write("# Automatically generated by FSGS\n")
        f.write("\n[sdl]\n")
        f.write("output=opengl\n")

        if self.use_fullscreen():
            f.write("fullscreen=true\n")
        else:
            f.write("fullscreen=false\n")
        f.write("fullresolution={0}x{1}\n".format(*self.screen_size()))

        f.write("\n[render]\n")
        f.write("frameskip=0\n")
        f.write("aspect=true\n")
        # if self.use_stretching():
        #     # This option does not stretch, it merely does not correct
        #     # aspect for non-square pixels resolutions, e.g. 320x200.
        #     # f.write("aspect=false\n")
        #     # This custom environment variable however, does cause stretching.
        #     self.env["FSGS_STRETCH"] = "1"
        # else:
        #     # f.write("aspect=true\n")
        #     self.env["FSGS_STRETCH"] = "0"

        if self.effect() == self.SCALE2X_EFFECT:
            scaler = "advmame2x"
        elif self.effect() == self.HQ2X_EFFECT:
            scaler = "hq2x"
        # elif self.effect() == self.CRT_EFFECT:
        #     scaler = "rgb2x"  # Looks ugly!
        else:
            scaler = "normal2x"

        if scaler:
            f.write("scaler={}\n".format(scaler))

        f.write("\n[cpu]\n")
        cpu_core = "auto"
        if self.options[Option.DOSBOX_CPU_CORE]:
            cpu_core = self.options[Option.DOSBOX_CPU_CORE]
            cpu_core = cpu_core.lower().strip()
        if System.windows and System.x86_64:
            # Dynamic core crashes on Windows x86-64
            print("[DOS] Forcing normal cpu core on Windows x86-64")
            if cpu_core in ["auto", "dynamic"]:
                cpu_core = "normal"
        f.write("core={0}\n".format(cpu_core))
        cpu_cycles = "auto"
        if self.options[Option.DOSBOX_CPU_CPUTYPE]:
            f.write(
                "cputype={0}\n".format(self.options[Option.DOSBOX_CPU_CPUTYPE])
            )
        if self.options[Option.DOSBOX_CPU_CYCLES]:
            cpu_cycles = self.options[Option.DOSBOX_CPU_CYCLES]
            if cpu_cycles.startswith("max "):
                pass
            elif cpu_cycles in ["auto", "max"]:
                pass
            else:
                cpu_cycles = "fixed " + cpu_cycles
        f.write("cycles={0}\n".format(cpu_cycles))

        f.write("\n[dosbox]\n")
        if self.options[Option.DOSBOX_MACHINE]:
            f.write("machine={}\n".format(self.options[Option.DOSBOX_MACHINE]))
        if self.options[Option.DOSBOX_MEMSIZE]:
            f.write("memsize={}\n".format(self.options[Option.DOSBOX_MEMSIZE]))

        self.configure_gus(f)
        self.configure_midi(f)
        self.configure_sblaster(f)

        f.write("\n[autoexec]\n")
        if self.options[Option.AUTO_LOAD] != "0":
            f.write("@ECHO OFF\n")
        # for drive, drive_path in self.drives:
        #     pass
        for name in os.listdir(self.drives_dir.path):
            if name in "ABCDEFGHIJKLMNOPQRSTUVWXYZ":
                p = os.path.join(self.drives_dir.path, name, "CD", "IMAGE.CUE")
                if os.path.exists(p):
                    f.write('IMGMOUNT {0} "{1}" -t iso\n'.format(name, p))
                    continue
                p = os.path.join(self.drives_dir.path, name, "CD", "IMAGE.ISO")
                if os.path.exists(p):
                    f.write('IMGMOUNT {0} "{1}" -t iso\n'.format(name, p))
                    continue
                if name in "DEF":
                    f.write(
                        'MOUNT {0} "{1}" -t cdrom\n'.format(
                            name, os.path.join(self.drives_dir.path, name)
                        )
                    )
                    continue
                f.write(
                    'MOUNT {0} "{1}"\n'.format(
                        name, os.path.join(self.drives_dir.path, name)
                    )
                )
        f.write("C:\n")
        f.write("CLS\n")
        # for i in range(25):
        #     f.write("echo.\n")
        if self.options[Option.AUTO_LOAD] == "0":
            f.write(
                "@echo Auto-load is disabled. The following is the "
                "normal commands to start this game:\n"
            )
            f.write("@echo.\n")
        for command in self.options[Option.COMMAND].split(";"):
            command = command.strip()
            command = command.replace("$DRIVES", self.drives_dir.path)
            if not self.options[
                Option.AUTO_LOAD
            ] == "0" or command.lower().split(" ")[0].strip("@") in [
                "imgmount",
                "mount",
            ]:
                f.write("{0}\n".format(command))
            else:
                f.write("@echo {0}\n".format(command))
        if self.options[Option.AUTO_LOAD] == "0":
            f.write("@echo.\n")
        else:
            if not self.options[Option.AUTO_QUIT] == "0":
                f.write("exit\n")

        if System.windows:
            # We don't want to open the separate console window on windows.
            self.emulator.args.append("-noconsole")

    def configure_gus(self, f):
        f.write("\n[gus]\n")
        if self.options[Option.DOSBOX_GUS_GUS].lower() == "true":
            f.write("gus=true\n")
            f.write("ultradir=U:\\ULTRASND\n")
            ultrasnd_drive = os.path.join(self.drives_dir.path, "U")
            source_dir = os.path.join(
                FSGSDirectories.get_data_dir(), "ULTRASND"
            )
            dest_dir = os.path.join(ultrasnd_drive, "ULTRASND")
            if os.path.exists(source_dir):
                shutil.copytree(source_dir, dest_dir)
            else:
                # FIXME: ADD ULTRASND WARNING
                pass

    def configure_midi(self, f):
        f.write("\n[midi]\n")
        if True:
            f.write("mpu401=intelligent\n")
            # f.write("mpu401=uart\n")
        if System.windows:
            pass
        elif System.macos:
            pass
        else:
            f.write("mididevice=alsa\n")
            f.write("midiconfig=128:0\n")

    def configure_sblaster(self, f):
        f.write("\n[sblaster]\n")
        if self.options[Option.DOSBOX_SBLASTER_SBTYPE]:
            f.write(
                "sbtype={}\n".format(
                    self.options[Option.DOSBOX_SBLASTER_SBTYPE]
                )
            )
        if self.options[Option.DOSBOX_SBLASTER_SBBASE]:
            f.write(
                "sbbase={}\n".format(
                    self.options[Option.DOSBOX_SBLASTER_SBBASE]
                )
            )
        if self.options[Option.DOSBOX_SBLASTER_IRQ]:
            f.write(
                "irq={}\n".format(self.options[Option.DOSBOX_SBLASTER_IRQ])
            )
        if self.options[Option.DOSBOX_SBLASTER_OPLRATE]:
            f.write(
                "oplrate={}\n".format(
                    self.options[Option.DOSBOX_SBLASTER_OPLRATE]
                )
            )



import os

import fsui
from fsbc.paths import Paths
from fsgs.checksumtool import ChecksumTool
from fsgs.FSGSDirectories import FSGSDirectories
from fsgs.amiga.amiga import Amiga
from fsgs.context import fsgs
from launcher.cd_manager import CDManager
from launcher.floppy_manager import FloppyManager
from launcher.i18n import gettext
from launcher.launcher_config import LauncherConfig
from launcher.option import Option
from launcher.ui.IconButton import IconButton
from launcher.ui.LauncherFilePicker import LauncherFilePicker
from launcher.ui.behaviors.configbehavior import ConfigBehavior
from launcher.ui.behaviors.platformbehavior import (
    AMIGA_PLATFORMS,
    PlatformEnableBehavior,
)


class MediaListGroup(fsui.Group):
    def __init__(self, parent, cd_mode):
        fsui.Group.__init__(self, parent)
        self.layout = fsui.VerticalLayout()

        self.cd_mode = cd_mode
        if self.cd_mode:
            self.file_key_prefix = "cdrom_image_"
            self.file_key = "cdrom_image_{0}"
            self.sha1_key = "x_cdrom_image_{0}_sha1"
            platforms = AMIGA_PLATFORMS
        else:
            self.file_key_prefix = "floppy_image_"
            self.file_key = "floppy_image_{0}"
            self.sha1_key = "x_floppy_image_{0}_sha1"
            platforms = AMIGA_PLATFORMS

        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, expand=False, fill=True)
        self.heading_label = fsui.HeadingLabel(
            self, gettext("Media Swap List")
        )
        hori_layout.add(
            self.heading_label, margin=10, margin_top=20, margin_bottom=20
        )
        hori_layout.add_spacer(0, expand=True)

        if not self.cd_mode:
            save_disk_check_box = SaveDiskCheckBox(self)
            hori_layout.add(save_disk_check_box, margin_right=20)
            PlatformEnableBehavior(save_disk_check_box, platforms=platforms)

        clear_button = IconButton(self, "clear_button.png")
        clear_button.set_tooltip(gettext("Clear List"))
        clear_button.activated.connect(self.on_clear_list)
        PlatformEnableBehavior(clear_button, platforms=platforms)
        hori_layout.add(clear_button, margin_right=10)
        remove_button = IconButton(self, "remove_button.png")
        remove_button.set_tooltip(gettext("Remove Selected Files"))
        remove_button.activated.connect(self.on_remove_button)
        PlatformEnableBehavior(remove_button, platforms=platforms)
        hori_layout.add(remove_button, margin_right=10)
        add_button = IconButton(self, "add_button.png")
        add_button.set_tooltip(gettext("Add Files to List"))
        add_button.activated.connect(self.on_add_button)
        PlatformEnableBehavior(add_button, platforms=platforms)
        hori_layout.add(add_button, margin_right=10)

        # hori_layout = fsui.HorizontalLayout()
        # self.layout.add(hori_layout, expand=True, fill=True)
        self.list_view = fsui.ListView(self)
        PlatformEnableBehavior(self.list_view, platforms=platforms)
        self.list_view.on_activate_item = self.on_activate_item
        if self.cd_mode:
            self.default_icon = fsui.Image("launcher:res/cdrom_16.png")
        else:
            self.default_icon = fsui.Image("launcher:res/floppy_16.png")
        # hori_layout.add(self.list_view, expand=True, fill=True, margin=10)
        self.layout.add(
            self.list_view, expand=True, fill=True, margin=10, margin_top=0
        )

        # vert_layout = fsui.VerticalLayout()
        # hori_layout.add(vert_layout, fill=True)

        # add_button = IconButton(self, "add_button.png")
        # add_button.set_tooltip(gettext("Add Files to List"))
        # add_button.activated.connect(self.on_add_button)
        # vert_layout.add(add_button, margin=10)
        #
        # remove_button = IconButton(self, "remove_button.png")
        # remove_button.set_tooltip(gettext("Remove Selected Files"))
        # remove_button.activated.connect(self.on_remove_button)
        # vert_layout.add(remove_button, margin=10)
        #
        # clear_button = IconButton(self, "clear_button.png")
        # clear_button.set_tooltip(gettext("Clear List"))
        # clear_button.activated.connect(self.on_clear_list)
        # vert_layout.add(clear_button, margin=10)

        self.update_list()
        LauncherConfig.add_listener(self)

    def on_destroy(self):
        LauncherConfig.remove_listener(self)

    def on_config(self, key, _):
        if key.startswith(self.file_key_prefix):
            self.update_list()

    def on_activate_item(self, item):
        path = LauncherConfig.get(self.file_key.format(item))
        sha1 = LauncherConfig.get(self.sha1_key.format(item))
        if self.cd_mode:
            pass
        else:
            fsgs.amiga.insert_floppy_in_free_drive(path, sha1=sha1)

    def create_list(self):
        items = []
        if self.cd_mode:
            max_items = Amiga.MAX_CDROM_IMAGES
        else:
            max_items = Amiga.MAX_FLOPPY_IMAGES
        for i in range(max_items):
            path = LauncherConfig.get(self.file_key.format(i))
            sha1 = LauncherConfig.get(self.sha1_key.format(i))
            if not path:
                continue
            items.append((path, sha1))
        return items

    def update_list(self):
        # items = []
        self.list_view.clear()
        for path, sha1 in self.create_list():
            dir_path, name = os.path.split(path)
            if dir_path:
                label = "{0} ({1})".format(name, dir_path)
            else:
                label = path
            self.list_view.add_item(label, self.default_icon)
            # self.list_view.set_items(items)

    def on_clear_list(self):
        if self.cd_mode:
            CDManager.clear_cdrom_list()
        else:
            FloppyManager.clear_floppy_list()

    def on_remove_button(self):
        index = self.list_view.get_index()
        existing_items = self.create_list()
        if 0 <= index < len(existing_items):
            del existing_items[index]
            self.set_new_config(existing_items)

    def on_add_button(self):
        existing_items = self.create_list()

        default_dir = FSGSDirectories.get_floppies_dir()
        if self.cd_mode:
            dialog = LauncherFilePicker(
                self.get_window(),
                gettext("Select Multiple CD-ROMs"),
                "cd",
                multiple=True,
            )
        else:
            dialog = LauncherFilePicker(
                self.get_window(),
                gettext("Select Multiple Floppies"),
                "floppy",
                multiple=True,
            )
        if not dialog.show_modal():
            print("dialog.show returned false")
            return
        print("dialog.show returned true")
        paths = dialog.get_paths()
        paths.sort()
        print(paths)

        checksum_tool = ChecksumTool(self.get_window())
        for i, path in enumerate(paths):
            sha1 = checksum_tool.checksum(path)
            path = Paths.contract_path(path, default_dir)

            dir_path, file = os.path.split(path)
            if os.path.normcase(
                os.path.normpath(dir_path)
            ) == os.path.normcase(os.path.normpath(default_dir)):
                path = file

            existing_items.append((path, sha1))
        self.set_new_config(existing_items)

    def set_new_config(self, items):
        if self.cd_mode:
            max_items = Amiga.MAX_CDROM_IMAGES
        else:
            max_items = Amiga.MAX_FLOPPY_IMAGES
        set_list = []
        for i in range(max(max_items, len(items))):
            if i >= max_items:
                break
            elif i >= len(items):
                path, sha1 = "", ""
            else:
                path, sha1 = items[i]
            set_list.append((self.file_key.format(i), path))
            set_list.append((self.sha1_key.format(i), sha1))
        LauncherConfig.set_multiple(set_list)


class SaveDiskCheckBox(fsui.CheckBox):
    def __init__(self, parent):
        super().__init__(parent, gettext("Include Save Disk"))
        self.set_tooltip(
            gettext(
                "When checked, include a save disk in FS-UAE's floppy swap list"
            )
        )
        ConfigBehavior(self, [Option.SAVE_DISK])

    def on_changed(self):
        LauncherConfig.set(Option.SAVE_DISK, "" if self.is_checked() else "0")

    def on_save_disk_config(self, value):
        self.check(value != "0")



# FSGS - Common functionality for FS Game System.
# Copyright (C) 2013-2019  Frode Solheim <frode@solheim.dev>
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.



import os

from fsbc.paths import Paths
from fsgs.archive import Archive
from fsgs.BaseContext import BaseContext
from fsgs.checksumtool import ChecksumTool
from fsgs.FSGSDirectories import FSGSDirectories
from fsgs.amiga.amiga import Amiga


class AmigaContext(BaseContext):
    def __init__(self, main_context):
        BaseContext.__init__(self, main_context)

    def reset_all_floppies(self):
        # FIXME: MAX_AMIGA_FLOPPY_DRIVES
        for i in range(4):
            self.eject_floppy(i)
        self.clear_floppy_list()

    def eject_floppy(self, drive):
        values = [
            ("floppy_drive_{0}".format(drive), ""),
            ("x_floppy_drive_{0}_sha1".format(drive), ""),
        ]
        self.set_config(values)

        # self.notify("fsgs:config:floppy_drive_{0}".format(drive), "")
        # self.notify("fsgs:config:x_floppy_drive_{0}_sha1".format(drive), "")

    def clear_floppy_list(self):
        values = []
        for i in range(Amiga.MAX_FLOPPY_IMAGES):
            values.append(("floppy_image_{0}".format(i), ""))
            values.append(("x_floppy_image_{0}_sha1".format(i), ""))
        self.set_config(values)

    def insert_floppy_in_free_drive(self, path, sha1=None):
        # FIXME: max floppy drives
        for i in range(4):
            if not self.config.get("floppy_drive_{0}".format(i)):
                self.insert_floppy(i, path, sha1=sha1)
                return True
        return False

    def insert_floppy(self, drive, path, sha1=None):
        if sha1 is None:
            sha1 = ChecksumTool().checksum(path)
        default_dir = FSGSDirectories.get_floppies_dir()
        path = Paths.contract_path(path, default_dir)
        self.set_config(
            [
                ("floppy_drive_{0}".format(drive), path),
                ("x_floppy_drive_{0}_sha1".format(drive), sha1),
            ]
        )

    def insert_cd(self, drive, path, sha1=None):
        if sha1 is None:
            sha1 = ""
            print("FIXME: not calculating CD checksum just yet")
        default_dir = FSGSDirectories.get_cdroms_dir()
        path = Paths.contract_path(path, default_dir)
        self.set_config(
            [
                ("cdrom_drive_{0}".format(drive), path),
                ("x_cdrom_drive_{0}_sha1".format(drive), sha1),
            ]
        )

    def insert_multiple_floppies(self, insert_paths):
        paths = []
        for path in insert_paths:
            embedded_files = []
            if path.endswith(".zip"):
                archive = Archive(path)
                files = archive.list_files()
                for file in files:
                    name, ext = os.path.splitext(file)
                    # FIXME: get list of floppy extensions from a central
                    # place
                    if ext in [".adf", ".ipf"]:
                        embedded_files.append(file)
            if len(embedded_files) > 0:
                embedded_files.sort()
                print("found embedded floppy images:")
                print(embedded_files)
                for file in embedded_files:
                    paths.append(file)
            else:
                paths.append(path)

        default_dir = FSGSDirectories.get_floppies_dir()
        checksum_tool = ChecksumTool()
        for i, path in enumerate(paths):
            sha1 = checksum_tool.checksum(path)
            path = Paths.contract_path(path, default_dir)

            if i < 4:
                self.set_config(
                    [
                        ("floppy_drive_{0}".format(i), path),
                        ("x_floppy_drive_{0}_sha1".format(i), sha1),
                    ]
                )
            self.set_config(
                [
                    ("floppy_image_{0}".format(i), path),
                    ("x_floppy_image_{0}_sha1".format(i), sha1),
                ]
            )

        # blank the rest of the drives
        for i in range(len(paths), 4):
            self.set_config(
                [
                    ("floppy_drive_{0}".format(i), ""),
                    ("x_floppy_drive_{0}_sha1".format(i), ""),
                ]
            )
        # blank the rest of the image list
        for i in range(len(paths), 20):
            self.set_config(
                [
                    ("floppy_image_{0}".format(i), ""),
                    ("x_floppy_image_{0}_sha1".format(i), ""),
                ]
            )

    def get_base_dir(self):
        # FIXME: dependency loop
        return FSGSDirectories.get_base_dir()

    def get_floppies_dir(self):
        # FIXME: dependency loop
        return FSGSDirectories.get_floppies_dir()

    def get_cdroms_dir(self):
        # FIXME: dependency loop
        return FSGSDirectories.get_cdroms_dir()

    def get_hard_drives_dir(self):
        # FIXME: dependency loop
        return FSGSDirectories.get_hard_drives_dir()

    def get_kickstarts_dir(self):
        # FIXME: dependency loop
        return FSGSDirectories.get_kickstarts_dir()

    def get_controllers_dir(self):
        # FIXME: dependency loop
        return FSGSDirectories.get_controllers_dir()



import fsui
from fsgs.option import Option
from launcher.launcher_settings import LauncherSettings


class WorkspaceTheme:

    __instance = None

    @classmethod
    def instance(cls):
        if not cls.__instance:
            cls.__instance = cls()
        return cls.__instance

    def __init__(self):
        self.title_font = fsui.Font("Noto Sans", 15)
        # self.title_color = fsui.Color(0x80, 0x80, 0x80)
        self.title_color = fsui.Color(0x44, 0x44, 0x44)
        self.title_background = fsui.Color(0xFF, 0xFF, 0xFF)
        # self.title_separator_color = fsui.Color(0xe5, 0xe5, 0xe5)
        self.title_separator_color = fsui.Color(0xCC, 0xCC, 0xCC)
        self.window_background = fsui.Color(0xF2, 0xF2, 0xF2)
        # self.sidebar_background = fsui.Color(0xeb, 0xeb, 0xeb)
        self.sidebar_background = fsui.Color(0xE2, 0xE2, 0xE2)
        self.selection_background = fsui.Color(0x40, 0x80, 0xFF)

        self.title_glow_color = None
        # self.title_glow_color = fsui.Color(0xff, 0xcc, 0xff, 0x80)

    @property
    def has_close_buttons(self):
        return LauncherSettings.get(Option.LAUNCHER_CLOSE_BUTTONS) == "1"
